#
# Created by: Pearu Peterson, April 2002
#

import math
import pytest
import numpy as np
from numpy.testing import (assert_equal, assert_almost_equal,
                           assert_array_almost_equal, assert_allclose)
from pytest import raises as assert_raises

from numpy import (arange, triu, tril, zeros, tril_indices, ones,
                   diag, append, eye, nonzero)

import scipy
from scipy.linalg import get_blas_funcs, toeplitz, solve
from scipy.linalg.blas import HAS_ILP64

try:
    from scipy.linalg import _fblas as fblas
    FBLAS_ERROR = fblas.__fblas_error
    HAS_LP64_FBLAS = True
except ImportError:
    fblas = None
    HAS_LP64_FBLAS = False

try:
    from scipy.linalg import _fblas_64 as fblas_64
    FBLAS_ERROR = fblas_64.__fblas_64_error
except ImportError:
    fblas_64 = None


REAL_DTYPES = [np.float32, np.float64]
COMPLEX_DTYPES = [np.complex64, np.complex128]
DTYPES = REAL_DTYPES + COMPLEX_DTYPES


def test_get_blas_funcs():
    # check that it returns Fortran code for arrays that are
    # fortran-ordered
    f1, f2, f3 = get_blas_funcs(
        ('axpy', 'axpy', 'axpy'),
        (np.empty((2, 2), dtype=np.complex64, order='F'),
         np.empty((2, 2), dtype=np.complex128, order='C'))
        )

    # get_blas_funcs will choose libraries depending on most generic
    # array
    assert_equal(f1.typecode, 'z')
    assert_equal(f2.typecode, 'z')

    # check defaults.
    f1 = get_blas_funcs('rotg')
    assert_equal(f1.typecode, 'd')

    # check also dtype interface
    f1 = get_blas_funcs('gemm', dtype=np.complex64)
    assert_equal(f1.typecode, 'c')
    f1 = get_blas_funcs('gemm', dtype='F')
    assert_equal(f1.typecode, 'c')

    # extended precision complex
    f1 = get_blas_funcs('gemm', dtype=np.clongdouble)
    assert_equal(f1.typecode, 'z')

    # check safe complex upcasting
    f1 = get_blas_funcs('axpy',
                        (np.empty((2, 2), dtype=np.float64),
                         np.empty((2, 2), dtype=np.complex64))
                        )
    assert_equal(f1.typecode, 'z')


def test_get_blas_funcs_ilp_preferred():
    # "preferred" mean ILP64 if available LP64 otherwise
    gemm = get_blas_funcs('gemm', (np.eye(3),), ilp64="preferred")
    assert gemm.int_dtype == np.int64 if HAS_ILP64 else np.int32
    assert gemm.module_name == 'fblas_64' if HAS_ILP64 else 'fblas'

    # default is "preferred"
    gemm = get_blas_funcs('gemm', (np.eye(3),))
    assert gemm.int_dtype == np.int64 if HAS_ILP64 else np.int32
    assert gemm.module_name == 'fblas_64' if HAS_ILP64 else 'fblas'


def test_get_blas_funcs_ilp_true():
    # True is ILP64 or fail if not available
    if HAS_ILP64:
        gemm = get_blas_funcs('gemm', (np.eye(3),), ilp64=True)
        assert gemm.int_dtype == np.int64
        assert gemm.module_name == 'fblas_64'
    else:
        with pytest.raises(RuntimeError):
            get_blas_funcs('gemm', (np.eye(3),), ilp64=True)


def test_get_blas_funcs_ilp_false():
    # False is LP64 or fail if not available
    if HAS_LP64_FBLAS:
        gemm = get_blas_funcs('gemm', (np.eye(3),), ilp64=False)
        assert gemm.int_dtype == np.int32
        assert gemm.module_name == 'fblas'
    else:
        with pytest.raises(ValueError):
            get_blas_funcs('gemm', (np.eye(3),), ilp64=False)


def test_get_blas_funcs_alias():
    # check alias for get_blas_funcs
    f, g = get_blas_funcs(('nrm2', 'dot'), dtype=np.complex64)
    assert f.typecode == 'c'
    assert g.typecode == 'c'

    f, g, h = get_blas_funcs(('dot', 'dotc', 'dotu'), dtype=np.float64)
    assert f is g
    assert f is h


def _dt_from_prefix(prefix):
    """Array dtype from a blas-style prefix."""
    if prefix.startswith('i'):
        prefix = prefix[1:]   # isamax, idamax etc
    elif prefix in ['sc', 'dz']:
        prefix = prefix[0]   # scasum, dzasum
    elif prefix in ['cs', 'zd']:
        prefix = prefix[0]  # zdscal, csscal

    dt_map = {'z': np.complex128, 'c': np.complex64, 'd': np.float64, 's': np.float32}
    return dt_map[prefix]


def parametrize_blas(func_name, prefixes, modules=None):
    """Parametrize a test over BLAS prefixes, "sdcz", and over the BLAS modules,
    `fblas,fblas_64.

    Given a func_name "gemm", this generates a pytest parametrization over up to 8
    variants: sgemm, dgemm, cgemm, zgemm, each loaded from `fblas` (i.e. 32-bit LP64
    variants) and `fblas_64` (i.e. 64-bit ILP64 variants)..

    If a module is not available, the pytest parameter has a skip mark, so that the
    test is skipped with a descriptive message.

    Parameters
    ----------
    func_name : str
        The base name of a BLAS function, e.g. "gemm"
    prefixes : str or sequence
        BLAS prefixes to prepend to `func_name`.
        E.g. ``func_name='gemm', prefixes='cz'`` generates `cgemm` and `zgemm`.
    modules : sequence of ``(module, str)`` pairs
        BLAS modules to fetch functions from. By default, use `fblas` (LP64 variant)
        and fblas_64 (ILP64 variant).
    """
    if modules is None:
        modules = [(fblas, "fblas"), (fblas_64, "fblas_64")]

    params = []
    for mod, mod_name in modules:
        for prefix in prefixes:
            dtype = _dt_from_prefix(prefix)
            if mod is None:
                param_ = pytest.param(
                    None, dtype,
                    id=f"{mod_name}.{prefix}{func_name}",
                    marks=pytest.mark.skip(reason=f"{mod_name} is not available")
                )
            else:
                # Fetch the BLAS function from the BLAS module. NB: if the name is not
                # found in the module, it's a hard failure (all names must be present).
                f = getattr(mod, prefix + func_name)
                param_ = pytest.param(f, dtype, id=f"{mod_name}.{prefix}{func_name}")

            params.append(param_)

    return pytest.mark.parametrize("f, dtype", params)


class TestFBLAS1Simple:

    @parametrize_blas("axpy", "sdcz")
    def test_axpy(self, f, dtype):
        assert_array_almost_equal(f([1, 2, 3], [2, -1, 3], a=5),
                                  [7, 9, 18])
        if dtype in COMPLEX_DTYPES:
            assert_array_almost_equal(f([1, 2j, 3], [2, -1, 3], a=5),
                                      [7, 10j-1, 18])

    @parametrize_blas("copy", "sdcz")
    def test_copy(self, f, dtype):
        assert_array_almost_equal(f([3, 4, 5], [8]*3), [3, 4, 5])
        if dtype in COMPLEX_DTYPES:
            assert_array_almost_equal(f([3, 4j, 5+3j], [8]*3), [3, 4j, 5+3j])

    @parametrize_blas("asum", ["s", "d", "sc", "dz"])
    def test_asum(self, f, dtype):
        assert_almost_equal(f([3, -4, 5]), 12)
        if dtype in COMPLEX_DTYPES:
            assert_almost_equal(f([3j, -4, 3-4j]), 14)

    @parametrize_blas("dot", "sd")
    def test_dot(self, f, dtype):
        assert_almost_equal(f([3, -4, 5], [2, 5, 1]), -9)

    @parametrize_blas("dotu", "cz")
    def test_dotu(self, f, dtype):
        assert_almost_equal(f([3j, -4, 3-4j], [2, 3, 1]), -9+2j)

    @parametrize_blas("dotc", "cz")
    def test_dotc(self, f, dtype):
        assert_almost_equal(f([3j, -4, 3-4j], [2, 3j, 1]), 3-14j)

    @parametrize_blas("nrm2", ["s", "d", "sc", "dz"])
    def test_nrm2(self, f, dtype):
        assert_almost_equal(f([3, -4, 5]), math.sqrt(50))
        if dtype in COMPLEX_DTYPES:
            assert_almost_equal(f([3j, -4, 3-4j]), math.sqrt(50))

    @parametrize_blas("scal", ["s", "d", "cs", "zd"])
    def test_scal(self, f, dtype):
        assert_array_almost_equal(f(2, [3, -4, 5]), [6, -8, 10])
        if dtype in COMPLEX_DTYPES:
            assert_array_almost_equal(f(3, [3j, -4, 3-4j]), [9j, -12, 9-12j])

    @parametrize_blas("swap", "sdcz")
    def test_swap(self, f, dtype):
        x, y = [2, 3, 1], [-2, 3, 7]
        x1, y1 = f(x, y)
        assert_array_almost_equal(x1, y)
        assert_array_almost_equal(y1, x)

        if dtype in COMPLEX_DTYPES:
            x, y = [2, 3j, 1], [-2, 3, 7-3j]
            x1, y1 = f(x, y)
            assert_array_almost_equal(x1, y)
            assert_array_almost_equal(y1, x)

    @parametrize_blas("amax", ["is", "id", "ic", "iz"])
    def test_amax(self, f, dtype):
        assert_equal(f([-2, 4, 3]), 1)
        if dtype in COMPLEX_DTYPES:
            assert_equal(f([-5, 4+3j, 6]), 1)

    # XXX: need tests for rot,rotm,rotg,rotmg


class TestFBLAS2Simple:
    @parametrize_blas("gemv", "sdcz")
    def test_gemv(self, f, dtype):
        assert_array_almost_equal(f(3, [[3]], [-4]), [-36])
        assert_array_almost_equal(f(3, [[3]], [-4], 3, [5]), [-21])
        if dtype in COMPLEX_DTYPES:
            assert_array_almost_equal(f(3j, [[3-4j]], [-4]), [-48-36j])
            assert_array_almost_equal(f(3j, [[3-4j]], [-4], 3, [5j]),
                                      [-48-21j])

    @parametrize_blas("ger", "sd")
    def test_ger(self, f, dtype):
        assert_array_almost_equal(f(1, [1, 2], [3, 4]), [[3, 4], [6, 8]])
        assert_array_almost_equal(f(2, [1, 2, 3], [3, 4]),
                                  [[6, 8], [12, 16], [18, 24]])
        assert_array_almost_equal(f(1, [1, 2], [3, 4],
                                  a=[[1, 2], [3, 4]]), [[4, 6], [9, 12]])
        if dtype in COMPLEX_DTYPES:
            assert_array_almost_equal(f(1, [1j, 2], [3, 4]),
                                      [[3j, 4j], [6, 8]])
            assert_array_almost_equal(f(2, [1j, 2j, 3j], [3j, 4j]),
                                      [[6, 8], [12, 16], [18, 24]])

    @parametrize_blas("geru", "cz")
    def test_geru(self, f, dtype):
        assert_array_almost_equal(f(1, [1j, 2], [3, 4]),
                                  [[3j, 4j], [6, 8]])
        assert_array_almost_equal(f(-2, [1j, 2j, 3j], [3j, 4j]),
                                  [[6, 8], [12, 16], [18, 24]])

    @parametrize_blas("gerc", "cz")
    def test_gerc(self, f, dtype):
        assert_array_almost_equal(f(1, [1j, 2], [3, 4]),
                                  [[3j, 4j], [6, 8]])
        assert_array_almost_equal(f(2, [1j, 2j, 3j], [3j, 4j]),
                                  [[6, 8], [12, 16], [18, 24]])

    @parametrize_blas("syr", "sdcz")
    def test_syr(self, f, dtype):
        x = np.arange(1, 5, dtype='d')
        resx = np.triu(x[:, np.newaxis] * x)
        resx_reverse = np.triu(x[::-1, np.newaxis] * x[::-1])
        y = np.linspace(0, 8.5, 17, endpoint=False)
        z = np.arange(1, 9, dtype='d').view('D')
        resz = np.triu(z[:, np.newaxis] * z)
        resz_reverse = np.triu(z[::-1, np.newaxis] * z[::-1])
        w = np.c_[np.zeros(4), z, np.zeros(4)].ravel()

        rtol = np.finfo(dtype).eps

        assert_allclose(f(1.0, x), resx, rtol=rtol)
        assert_allclose(f(1.0, x, lower=True), resx.T, rtol=rtol)
        assert_allclose(f(1.0, y, incx=2, offx=2, n=4), resx, rtol=rtol)
        # negative increments imply reversed vectors in blas
        assert_allclose(f(1.0, y, incx=-2, offx=2, n=4),
                        resx_reverse, rtol=rtol)

        if dtype in COMPLEX_DTYPES:
            assert_allclose(f(1.0, z), resz, rtol=rtol)
            assert_allclose(f(1.0, z, lower=True), resz.T, rtol=rtol)
            assert_allclose(f(1.0, w, incx=3, offx=1, n=4), resz, rtol=rtol)
            # negative increments imply reversed vectors in blas
            assert_allclose(f(1.0, w, incx=-3, offx=1, n=4),
                            resz_reverse, rtol=rtol)

            a = np.zeros((4, 4), dtype, 'F')
            b = f(1.0, z, a=a, overwrite_a=True)
            assert_allclose(a, resz, rtol=rtol)
            b = f(2.0, z, a=a)
            assert a is not b
            assert_allclose(b, 3*resz, rtol=rtol)

        else:
            a = np.zeros((4, 4), dtype, 'F')
            b = f(1.0, x, a=a, overwrite_a=True)
            assert_allclose(a, resx, rtol=rtol)
            b = f(2.0, x, a=a)
            assert a is not b
            assert_allclose(b, 3*resx, rtol=rtol)

        assert_raises(Exception, f, 1.0, x, incx=0)
        assert_raises(Exception, f, 1.0, x, offx=5)
        assert_raises(Exception, f, 1.0, x, offx=-2)
        assert_raises(Exception, f, 1.0, x, n=-2)
        assert_raises(Exception, f, 1.0, x, n=5)
        assert_raises(Exception, f, 1.0, x, lower=2)
        assert_raises(Exception, f, 1.0, x, a=np.zeros((2, 2), 'd', 'F'))

    @parametrize_blas("her", "cz")
    def test_her(self, f, dtype):
        x = np.arange(1, 5, dtype='d')
        z = np.arange(1, 9, dtype='d').view('D')
        rehz = np.triu(z[:, np.newaxis] * z.conj())
        rehz_reverse = np.triu(z[::-1, np.newaxis] * z[::-1].conj())
        w = np.c_[np.zeros(4), z, np.zeros(4)].ravel()

        rtol = np.finfo(dtype).eps

        assert_allclose(f(1.0, z), rehz, rtol=rtol)
        assert_allclose(f(1.0, z, lower=True), rehz.T.conj(), rtol=rtol)
        assert_allclose(f(1.0, w, incx=3, offx=1, n=4), rehz, rtol=rtol)
        # negative increments imply reversed vectors in blas
        assert_allclose(f(1.0, w, incx=-3, offx=1, n=4),
                        rehz_reverse, rtol=rtol)

        a = np.zeros((4, 4), dtype, 'F')
        b = f(1.0, z, a=a, overwrite_a=True)
        assert_allclose(a, rehz, rtol=rtol)

        b = f(2.0, z, a=a)
        assert a is not b
        assert_allclose(b, 3*rehz, rtol=rtol)

        assert_raises(Exception, f, 1.0, x, incx=0)
        assert_raises(Exception, f, 1.0, x, offx=5)
        assert_raises(Exception, f, 1.0, x, offx=-2)
        assert_raises(Exception, f, 1.0, x, n=-2)
        assert_raises(Exception, f, 1.0, x, n=5)
        assert_raises(Exception, f, 1.0, x, lower=2)
        assert_raises(Exception, f, 1.0, x, a=np.zeros((2, 2), 'd', 'F'))

    @parametrize_blas("syr2", "sd")
    def test_syr2(self, f, dtype):
        x = np.arange(1, 5, dtype='d')
        y = np.arange(5, 9, dtype='d')
        resxy = np.triu(x[:, np.newaxis] * y + y[:, np.newaxis] * x)
        resxy_reverse = np.triu(x[::-1, np.newaxis] * y[::-1]
                                + y[::-1, np.newaxis] * x[::-1])

        q = np.linspace(0, 8.5, 17, endpoint=False)
        rtol = np.finfo(dtype).eps

        assert_allclose(f(1.0, x, y), resxy, rtol=rtol)
        assert_allclose(f(1.0, x, y, n=3), resxy[:3, :3], rtol=rtol)
        assert_allclose(f(1.0, x, y, lower=True), resxy.T, rtol=rtol)

        assert_allclose(f(1.0, q, q, incx=2, offx=2, incy=2, offy=10),
                        resxy, rtol=rtol)
        assert_allclose(f(1.0, q, q, incx=2, offx=2, incy=2, offy=10, n=3),
                        resxy[:3, :3], rtol=rtol)
        # negative increments imply reversed vectors in blas
        assert_allclose(f(1.0, q, q, incx=-2, offx=2, incy=-2, offy=10),
                        resxy_reverse, rtol=rtol)

        a = np.zeros((4, 4), dtype, 'F')
        b = f(1.0, x, y, a=a, overwrite_a=True)
        assert_allclose(a, resxy, rtol=rtol)

        b = f(2.0, x, y, a=a)
        assert a is not b
        assert_allclose(b, 3*resxy, rtol=rtol)

        assert_raises(Exception, f, 1.0, x, y, incx=0)
        assert_raises(Exception, f, 1.0, x, y, offx=5)
        assert_raises(Exception, f, 1.0, x, y, offx=-2)
        assert_raises(Exception, f, 1.0, x, y, incy=0)
        assert_raises(Exception, f, 1.0, x, y, offy=5)
        assert_raises(Exception, f, 1.0, x, y, offy=-2)
        assert_raises(Exception, f, 1.0, x, y, n=-2)
        assert_raises(Exception, f, 1.0, x, y, n=5)
        assert_raises(Exception, f, 1.0, x, y, lower=2)
        assert_raises(Exception, f, 1.0, x, y, a=np.zeros((2, 2), 'd', 'F'))

    @parametrize_blas("her2", "cz")
    def test_her2(self, f, dtype):
        x = np.arange(1, 9, dtype='d').view('D')
        y = np.arange(9, 17, dtype='d').view('D')
        resxy = x[:, np.newaxis] * y.conj() + y[:, np.newaxis] * x.conj()
        resxy = np.triu(resxy)

        resxy_reverse = x[::-1, np.newaxis] * y[::-1].conj()
        resxy_reverse += y[::-1, np.newaxis] * x[::-1].conj()
        resxy_reverse = np.triu(resxy_reverse)

        u = np.c_[np.zeros(4), x, np.zeros(4)].ravel()
        v = np.c_[np.zeros(4), y, np.zeros(4)].ravel()

        rtol = np.finfo(dtype).eps

        assert_allclose(f(1.0, x, y), resxy, rtol=rtol)
        assert_allclose(f(1.0, x, y, n=3), resxy[:3, :3], rtol=rtol)
        assert_allclose(f(1.0, x, y, lower=True), resxy.T.conj(),
                        rtol=rtol)

        assert_allclose(f(1.0, u, v, incx=3, offx=1, incy=3, offy=1),
                        resxy, rtol=rtol)
        assert_allclose(f(1.0, u, v, incx=3, offx=1, incy=3, offy=1, n=3),
                        resxy[:3, :3], rtol=rtol)
        # negative increments imply reversed vectors in blas
        assert_allclose(f(1.0, u, v, incx=-3, offx=1, incy=-3, offy=1),
                        resxy_reverse, rtol=rtol)

        a = np.zeros((4, 4), dtype, 'F')
        b = f(1.0, x, y, a=a, overwrite_a=True)
        assert_allclose(a, resxy, rtol=rtol)

        b = f(2.0, x, y, a=a)
        assert a is not b
        assert_allclose(b, 3*resxy, rtol=rtol)

        assert_raises(Exception, f, 1.0, x, y, incx=0)
        assert_raises(Exception, f, 1.0, x, y, offx=5)
        assert_raises(Exception, f, 1.0, x, y, offx=-2)
        assert_raises(Exception, f, 1.0, x, y, incy=0)
        assert_raises(Exception, f, 1.0, x, y, offy=5)
        assert_raises(Exception, f, 1.0, x, y, offy=-2)
        assert_raises(Exception, f, 1.0, x, y, n=-2)
        assert_raises(Exception, f, 1.0, x, y, n=5)
        assert_raises(Exception, f, 1.0, x, y, lower=2)
        assert_raises(Exception, f, 1.0, x, y,
                      a=np.zeros((2, 2), 'd', 'F'))

    @parametrize_blas("gbmv", "sdcz")
    def test_gbmv(self, f, dtype):
        rng = np.random.default_rng(1234)
        n = 7
        m = 5
        kl = 1
        ku = 2
        # fake a banded matrix via toeplitz
        A = toeplitz(append(rng.random(kl+1), zeros(m-kl-1)),
                     append(rng.random(ku+1), zeros(n-ku-1)))
        A = A.astype(dtype)
        Ab = zeros((kl+ku+1, n), dtype=dtype)

        # Form the banded storage
        Ab[2, :5] = A[0, 0]  # diag
        Ab[1, 1:6] = A[0, 1]  # sup1
        Ab[0, 2:7] = A[0, 2]  # sup2
        Ab[3, :4] = A[1, 0]  # sub1

        x = rng.random(n).astype(dtype)
        y = rng.random(m).astype(dtype)
        alpha, beta = dtype(3), dtype(-5)

        func = f
        y1 = func(m=m, n=n, ku=ku, kl=kl, alpha=alpha, a=Ab,
                  x=x, y=y, beta=beta)
        y2 = alpha * A.dot(x) + beta * y
        assert_array_almost_equal(y1, y2)

        y1 = func(m=m, n=n, ku=ku, kl=kl, alpha=alpha, a=Ab,
                  x=y, y=x, beta=beta, trans=1)
        y2 = alpha * A.T.dot(y) + beta * x
        assert_array_almost_equal(y1, y2)

    @pytest.mark.parametrize("ilp64", [True, False])
    @pytest.mark.parametrize("dtype", DTYPES)
    def test_sbmv_hbmv(self, dtype, ilp64):
        mod = fblas_64 if ilp64 else fblas
        if mod is None:
            mod_name = "fblas_64" if ilp64 else "fblas"
            pytest.skip(f"{mod_name} is not available")

        rng = np.random.default_rng(1234)
        n = 6
        k = 2
        A = zeros((n, n), dtype=dtype)
        Ab = zeros((k+1, n), dtype=dtype)

        # Form the array and its packed banded storage
        A[arange(n), arange(n)] = rng.random(n)
        for ind2 in range(1, k+1):
            temp = rng.random(n-ind2)
            A[arange(n-ind2), arange(ind2, n)] = temp
            Ab[-1-ind2, ind2:] = temp
        A = A.astype(dtype)
        if dtype in COMPLEX_DTYPES:
            A += A.conj().T
            func, = get_blas_funcs(('hbmv',), dtype=dtype, ilp64="preferred")
        else:
            A += A.T
            func, = get_blas_funcs(('sbmv',), dtype=dtype, ilp64="preferred")

        Ab[-1, :] = diag(A)
        x = rng.random(n).astype(dtype)
        y = rng.random(n).astype(dtype)
        alpha, beta = dtype(1.25), dtype(3)

        y1 = func(k=k, alpha=alpha, a=Ab, x=x, y=y, beta=beta)
        y2 = alpha * A.dot(x) + beta * y
        assert_array_almost_equal(y1, y2)

    @pytest.mark.parametrize("ilp64", [True, False])
    @pytest.mark.parametrize("fname,dtype", [
        *[('spmv', dtype) for dtype in REAL_DTYPES + COMPLEX_DTYPES],
        *[('hpmv', dtype) for dtype in COMPLEX_DTYPES],
    ])
    def test_spmv_hpmv(self, fname, dtype, ilp64):
        mod = fblas_64 if ilp64 else fblas
        if mod is None:
            mod_name = "fblas_64" if ilp64 else "fblas"
            pytest.skip(f"{mod_name} is not available")

        rng = np.random.default_rng(1234)
        n = 3
        A = rng.random((n, n)).astype(dtype)
        if dtype in COMPLEX_DTYPES:
            A += rng.random((n, n))*1j
        A += A.T if fname == 'spmv' else A.conj().T
        c, r = tril_indices(n)
        Ap = A[r, c]
        x = rng.random(n).astype(dtype)
        y = rng.random(n).astype(dtype)
        xlong = arange(2*n).astype(dtype)
        ylong = ones(2*n).astype(dtype)
        alpha, beta = dtype(1.25), dtype(2)

        func, = get_blas_funcs((fname,), dtype=dtype, ilp64="preferred")
        y1 = func(n=n, alpha=alpha, ap=Ap, x=x, y=y, beta=beta)
        y2 = alpha * A.dot(x) + beta * y
        assert_array_almost_equal(y1, y2)

        # Test inc and offsets
        y1 = func(n=n-1, alpha=alpha, beta=beta, x=xlong, y=ylong, ap=Ap,
                  incx=2, incy=2, offx=n, offy=n)
        y2 = (alpha * A[:-1, :-1]).dot(xlong[3::2]) + beta * ylong[3::2]
        assert_array_almost_equal(y1[3::2], y2)
        assert_almost_equal(y1[4], ylong[4])

    @pytest.mark.parametrize("ilp64", [True, False])
    @pytest.mark.parametrize("fname,dtype", [
        *[('spr', dtype) for dtype in REAL_DTYPES + COMPLEX_DTYPES],
        *[('hpr', dtype) for dtype in COMPLEX_DTYPES],
    ])
    def test_spr_hpr(self, fname, dtype, ilp64):
        mod = fblas_64 if ilp64 else fblas
        if mod is None:
            mod_name = "fblas_64" if ilp64 else "fblas"
            pytest.skip(f"{mod_name} is not available")

        rng = np.random.default_rng(1234)
        n = 3
        A = rng.random((n, n)).astype(dtype)
        if dtype in COMPLEX_DTYPES:
            A += rng.random((n, n))*1j
        A += A.T if fname == 'spr' else A.conj().T
        c, r = tril_indices(n)
        Ap = A[r, c]
        x = rng.random(n).astype(dtype)

        alpha = np.finfo(dtype).dtype.type(2.5)
        if fname == 'hpr':
            func, = get_blas_funcs(('hpr',), dtype=dtype, ilp64="preferred")
            y2 = alpha * x[:, None].dot(x[None, :].conj()) + A
        else:
            func, = get_blas_funcs(('spr',), dtype=dtype, ilp64="preferred")
            y2 = alpha * x[:, None].dot(x[None, :]) + A

        y1 = func(n=n, alpha=alpha, ap=Ap, x=x)
        y1f = zeros((3, 3), dtype=dtype)
        y1f[r, c] = y1
        y1f[c, r] = y1.conj() if fname == 'hpr' else y1
        assert_array_almost_equal(y1f, y2)

    @pytest.mark.parametrize("ilp64", [True, False])
    @pytest.mark.parametrize("dtype", DTYPES)
    def test_spr2_hpr2(self, dtype, ilp64):
        mod = fblas_64 if ilp64 else fblas
        if mod is None:
            mod_name = "fblas_64" if ilp64 else "fblas"
            pytest.skip(f"{mod_name} is not available")

        rng = np.random.default_rng(1234)
        n = 3
        A = rng.random((n, n)).astype(dtype)
        if dtype in COMPLEX_DTYPES:
            A += rng.random((n, n))*1j
            A += A.conj().T
            func, = get_blas_funcs(('hpr2',), dtype=dtype, ilp64="preferred")
        else:
            A += A.T
            func, = get_blas_funcs(('spr2',), dtype=dtype, ilp64="preferred")

        c, r = tril_indices(n)
        Ap = A[r, c]
        x = rng.random(n).astype(dtype)
        y = rng.random(n).astype(dtype)
        alpha = dtype(2)

        u = alpha.conj() * x[:, None].dot(y[None, :].conj())
        y2 = A + u + u.conj().T
        y1 = func(n=n, alpha=alpha, x=x, y=y, ap=Ap)
        y1f = zeros((3, 3), dtype=dtype)
        y1f[r, c] = y1
        y1f[[1, 2, 2], [0, 0, 1]] = y1[[1, 3, 4]].conj()
        assert_array_almost_equal(y1f, y2)

    @parametrize_blas('tbmv', 'sdcz')
    def test_tbmv(self, f, dtype):
        rng = np.random.default_rng(1234)
        n = 10
        k = 3
        x = rng.random(n).astype(dtype)
        A = zeros((n, n), dtype=dtype)
        # Banded upper triangular array
        for sup in range(k+1):
            A[arange(n-sup), arange(sup, n)] = rng.random(n-sup)

        # Add complex parts for c,z
        if dtype in COMPLEX_DTYPES:
            A[nonzero(A)] += 1j * rng.random((k+1)*n-(k*(k+1)//2)).astype(dtype)

        # Form the banded storage
        Ab = zeros((k+1, n), dtype=dtype)
        for row in range(k+1):
            Ab[-row-1, row:] = diag(A, k=row)
        func = f

        y1 = func(k=k, a=Ab, x=x)
        y2 = A.dot(x)
        assert_array_almost_equal(y1, y2)

        y1 = func(k=k, a=Ab, x=x, diag=1)
        A[arange(n), arange(n)] = dtype(1)
        y2 = A.dot(x)
        assert_array_almost_equal(y1, y2)

        y1 = func(k=k, a=Ab, x=x, diag=1, trans=1)
        y2 = A.T.dot(x)
        assert_array_almost_equal(y1, y2)

        y1 = func(k=k, a=Ab, x=x, diag=1, trans=2)
        y2 = A.conj().T.dot(x)
        assert_array_almost_equal(y1, y2)

    @parametrize_blas('tbsv', 'sdcz')
    def test_tbsv(self, f, dtype):
        rng = np.random.default_rng(12345)
        n = 6
        k = 3
        x = rng.random(n).astype(dtype)
        A = zeros((n, n), dtype=dtype)
        # Banded upper triangular array
        for sup in range(k+1):
            A[arange(n-sup), arange(sup, n)] = rng.random(n-sup)

        # Add complex parts for c,z
        if dtype in COMPLEX_DTYPES:
            A[nonzero(A)] += 1j * rng.random((k+1)*n-(k*(k+1)//2)).astype(dtype)

        # Form the banded storage
        Ab = zeros((k+1, n), dtype=dtype)
        for row in range(k+1):
            Ab[-row-1, row:] = diag(A, k=row)
        func = f

        y1 = func(k=k, a=Ab, x=x)
        y2 = solve(A, x)
        assert_array_almost_equal(y1, y2)

        y1 = func(k=k, a=Ab, x=x, diag=1)
        A[arange(n), arange(n)] = dtype(1)
        y2 = solve(A, x)
        assert_array_almost_equal(y1, y2)

        y1 = func(k=k, a=Ab, x=x, diag=1, trans=1)
        y2 = solve(A.T, x)
        assert_array_almost_equal(y1, y2)

        y1 = func(k=k, a=Ab, x=x, diag=1, trans=2)
        y2 = solve(A.conj().T, x)
        assert_array_almost_equal(y1, y2)

    @parametrize_blas('tpmv', 'sdcz')
    def test_tpmv(self, f, dtype):
        rng = np.random.default_rng(1234)
        n = 10
        x = rng.random(n).astype(dtype)
        # Upper triangular array
        if dtype in COMPLEX_DTYPES:
            A = triu(rng.random((n, n)) + rng.random((n, n))*1j)
        else:
            A = triu(rng.random((n, n)))

        # Form the packed storage
        c, r = tril_indices(n)
        Ap = A[r, c]
        func = f

        y1 = func(n=n, ap=Ap, x=x)
        y2 = A.dot(x)
        assert_array_almost_equal(y1, y2)

        y1 = func(n=n, ap=Ap, x=x, diag=1)
        A[arange(n), arange(n)] = dtype(1)
        y2 = A.dot(x)
        assert_array_almost_equal(y1, y2)

        y1 = func(n=n, ap=Ap, x=x, diag=1, trans=1)
        y2 = A.T.dot(x)
        assert_array_almost_equal(y1, y2)

        y1 = func(n=n, ap=Ap, x=x, diag=1, trans=2)
        y2 = A.conj().T.dot(x)
        assert_array_almost_equal(y1, y2)

    @parametrize_blas('tpsv', 'sdcz')
    def test_tpsv(self, f, dtype):
        rng = np.random.default_rng(1234)
        n = 10
        x = rng.random(n).astype(dtype)
        # Upper triangular array
        if dtype in COMPLEX_DTYPES:
            A = triu(rng.random((n, n)) + rng.random((n, n))*1j)
        else:
            A = triu(rng.random((n, n)))
        A += eye(n)
        # Form the packed storage
        c, r = tril_indices(n)
        Ap = A[r, c]
        func = f

        y1 = func(n=n, ap=Ap, x=x)
        y2 = solve(A, x)
        assert_array_almost_equal(y1, y2)

        y1 = func(n=n, ap=Ap, x=x, diag=1)
        A[arange(n), arange(n)] = dtype(1)
        y2 = solve(A, x)
        assert_array_almost_equal(y1, y2)

        y1 = func(n=n, ap=Ap, x=x, diag=1, trans=1)
        y2 = solve(A.T, x)
        assert_array_almost_equal(y1, y2)

        y1 = func(n=n, ap=Ap, x=x, diag=1, trans=2)
        y2 = solve(A.conj().T, x)
        assert_array_almost_equal(y1, y2)

    @parametrize_blas('trmv', 'sdcz')
    def test_trmv(self, f, dtype):
        rng = np.random.default_rng(1234)
        n = 3
        A = (rng.random((n, n))+eye(n)).astype(dtype)
        x = rng.random(3).astype(dtype)
        func = f

        y1 = func(a=A, x=x)
        y2 = triu(A).dot(x)
        assert_array_almost_equal(y1, y2)

        y1 = func(a=A, x=x, diag=1)
        A[arange(n), arange(n)] = dtype(1)
        y2 = triu(A).dot(x)
        assert_array_almost_equal(y1, y2)

        y1 = func(a=A, x=x, diag=1, trans=1)
        y2 = triu(A).T.dot(x)
        assert_array_almost_equal(y1, y2)

        y1 = func(a=A, x=x, diag=1, trans=2)
        y2 = triu(A).conj().T.dot(x)
        assert_array_almost_equal(y1, y2)

    @parametrize_blas('trsv', 'sdcz')
    def test_trsv(self, f, dtype):
        rng = np.random.default_rng(1234)
        n = 15
        A = (rng.random((n, n))+eye(n)).astype(dtype)
        x = rng.random(n).astype(dtype)
        func = f

        y1 = func(a=A, x=x)
        y2 = solve(triu(A), x)
        assert_array_almost_equal(y1, y2)

        y1 = func(a=A, x=x, lower=1)
        y2 = solve(tril(A), x)
        assert_array_almost_equal(y1, y2)

        y1 = func(a=A, x=x, diag=1)
        A[arange(n), arange(n)] = dtype(1)
        y2 = solve(triu(A), x)
        assert_array_almost_equal(y1, y2)

        y1 = func(a=A, x=x, diag=1, trans=1)
        y2 = solve(triu(A).T, x)
        assert_array_almost_equal(y1, y2)

        y1 = func(a=A, x=x, diag=1, trans=2)
        y2 = solve(triu(A).conj().T, x)
        assert_array_almost_equal(y1, y2)


class TestFBLAS3Simple:
    @parametrize_blas("gemm", "sdcz")
    def test_gemm(self, f, dtype):
        assert_array_almost_equal(f(3, [3], [-4]), [[-36]])
        assert_array_almost_equal(f(3, [3], [-4], 3, [5]), [-21])
        if dtype in COMPLEX_DTYPES:
            assert_array_almost_equal(f(3j, [3-4j], [-4]), [[-48-36j]])
            assert_array_almost_equal(f(3j, [3-4j], [-4], 3, [5j]), [-48-21j])


class TestBLAS3Symm:

    def setup_method(self):
        self.a = np.array([[1., 2.],
                           [0., 1.]])
        self.b = np.array([[1., 0., 3.],
                           [0., -1., 2.]])
        self.c = np.ones((2, 3))
        self.t = np.array([[2., -1., 8.],
                           [3., 0., 9.]])

    @parametrize_blas("symm", "sdcz")
    def test_symm(self, f, dtype):
        res = f(a=self.a, b=self.b, c=self.c, alpha=1., beta=1.)
        assert_array_almost_equal(res, self.t)

        res = f(a=self.a.T, b=self.b, lower=1, c=self.c, alpha=1., beta=1.)
        assert_array_almost_equal(res, self.t)

        res = f(a=self.a, b=self.b.T, side=1, c=self.c.T,
                alpha=1., beta=1.)
        assert_array_almost_equal(res, self.t.T)

    @parametrize_blas("symm", "sdcz")
    def test_symm_wrong_side(self, f, dtype):
        """`side=1` means C <- B*A, hence shapes of A and B are to be
        compatible. Otherwise, f2py exception is raised.
        """
        # FIXME narrow down to _fblas.error
        with pytest.raises(Exception):
            f(a=self.a, b=self.b, alpha=1, side=1)

    @parametrize_blas("symm", "sdcz")
    def test_symm_wrong_uplo(self, f, dtype):
        """SYMM only considers the upper/lower part of A. Hence setting
        wrong value for `lower` (default is lower=0, meaning upper triangle)
        gives a wrong result.
        """
        res = f(a=self.a, b=self.b, c=self.c, alpha=1., beta=1.)
        assert np.allclose(res, self.t)
        res = f(a=self.a, b=self.b, lower=1, c=self.c, alpha=1., beta=1.)
        assert not np.allclose(res, self.t)


class TestBLAS3Syrk:
    def setup_method(self):
        self.a = np.array([[1., 0.],
                           [0., -2.],
                           [2., 3.]])
        self.t = np.array([[1., 0., 2.],
                           [0., 4., -6.],
                           [2., -6., 13.]])
        self.tt = np.array([[5., 6.],
                            [6., 13.]])

    @parametrize_blas("syrk", "sdcz")
    def test_syrk(self, f, dtype):
        c = f(a=self.a, alpha=1.)
        assert_array_almost_equal(np.triu(c), np.triu(self.t))

        c = f(a=self.a, alpha=1., lower=1)
        assert_array_almost_equal(np.tril(c), np.tril(self.t))

        c0 = np.ones(self.t.shape)
        c = f(a=self.a, alpha=1., beta=1., c=c0)
        assert_array_almost_equal(np.triu(c), np.triu(self.t+c0))

        c = f(a=self.a, alpha=1., trans=1)
        assert_array_almost_equal(np.triu(c), np.triu(self.tt))

    # prints '0-th dimension must be fixed to 3 but got 5',
    # FIXME: suppress?
    @parametrize_blas("syrk", "sdcz")
    def test_syrk_wrong_c(self, f, dtype):
        # FIXME narrow down to _fblas.error
        with pytest.raises(Exception):
            f(a=self.a, alpha=1., c=np.ones((5, 8)))
        # if C is supplied, it must have compatible dimensions


class TestBLAS3Syr2k:
    def setup_method(self):
        self.a = np.array([[1., 0.],
                           [0., -2.],
                           [2., 3.]])
        self.b = np.array([[0., 1.],
                           [1., 0.],
                           [0, 1.]])
        self.t = np.array([[0., -1., 3.],
                           [-1., 0., 0.],
                           [3., 0., 6.]])
        self.tt = np.array([[0., 1.],
                            [1., 6]])

    @parametrize_blas("syr2k", "sdcz")
    def test_syr2k(self, f, dtype):
        c = f(a=self.a, b=self.b, alpha=1.)
        assert_array_almost_equal(np.triu(c), np.triu(self.t))

        c = f(a=self.a, b=self.b, alpha=1., lower=1)
        assert_array_almost_equal(np.tril(c), np.tril(self.t))

        c0 = np.ones(self.t.shape)
        c = f(a=self.a, b=self.b, alpha=1., beta=1., c=c0)
        assert_array_almost_equal(np.triu(c), np.triu(self.t+c0))

        c = f(a=self.a, b=self.b, alpha=1., trans=1)
        assert_array_almost_equal(np.triu(c), np.triu(self.tt))

    # prints '0-th dimension must be fixed to 3 but got 5', FIXME: suppress?
    @parametrize_blas("syr2k", "sdcz")
    def test_syr2k_wrong_c(self, f, dtype):
        with pytest.raises(Exception):
            f(a=self.a, b=self.b, alpha=1., c=np.zeros((15, 8)))
        # if C is supplied, it must have compatible dimensions


class TestSyHe:
    """Quick and simple tests for (zc)-symm, syrk, syr2k."""

    def setup_method(self):
        self.sigma_y = np.array([[0., -1.j],
                                 [1.j, 0.]])

    @parametrize_blas("symm", "zc")
    def test_symm(self, f, dtype):
        # NB: a is symmetric w/upper diag of ONLY
        res = f(a=self.sigma_y, b=self.sigma_y, alpha=1.)
        assert_array_almost_equal(np.triu(res), np.diag([1, -1]))

    @parametrize_blas("hemm", "zc")
    def test_hemm(self, f, dtype):
        # NB: a is hermitian w/upper diag of ONLY
        res = f(a=self.sigma_y, b=self.sigma_y, alpha=1.)
        assert_array_almost_equal(np.triu(res), np.diag([1, 1]))

    @parametrize_blas("syrk", "zc")
    def test_syrk(self, f, dtype):
        res = f(a=self.sigma_y, alpha=1.)
        assert_array_almost_equal(np.triu(res), np.diag([-1, -1]))

    @parametrize_blas("herk", "zc")
    def test_herk(self, f, dtype):
        res = f(a=self.sigma_y, alpha=1.)
        assert_array_almost_equal(np.triu(res), np.diag([1, 1]))

    @parametrize_blas("syr2k", "zc")
    def test_syr2k_zr(self, f, dtype):
        res = f(a=self.sigma_y, b=self.sigma_y, alpha=1.)
        assert_array_almost_equal(np.triu(res), 2.*np.diag([-1, -1]))

    @parametrize_blas("her2k", "zc")
    def test_her2k_zr(self, f, dtype):
        res = f(a=self.sigma_y, b=self.sigma_y, alpha=1.)
        assert_array_almost_equal(np.triu(res), 2.*np.diag([1, 1]))


class TestTRMM:
    """Quick and simple tests for *trmm."""

    def setup_method(self):
        self.a = np.array([[1., 2., ],
                           [-2., 1.]])
        self.b = np.array([[3., 4., -1.],
                           [5., 6., -2.]])

        self.a2 = np.array([[1, 1, 2, 3],
                            [0, 1, 4, 5],
                            [0, 0, 1, 6],
                            [0, 0, 0, 1]], order="f")
        self.b2 = np.array([[1, 4], [2, 5], [3, 6], [7, 8], [9, 10]],
                           order="f")

    @parametrize_blas("trmm", "sdzc")
    def test_side(self, f, dtype):
        trmm = f
        # Provide large A array that works for side=1 but not 0 (see gh-10841)
        assert_raises(Exception, trmm, 1.0, self.a2, self.b2)
        res = trmm(1.0, self.a2.astype(dtype), self.b2.astype(dtype),
                   side=1)
        k = self.b2.shape[1]
        assert_allclose(res, self.b2 @ self.a2[:k, :k], rtol=0.,
                        atol=100*np.finfo(dtype).eps)

    @parametrize_blas("trmm", "sdcz")
    def test_ab(self, f, dtype):
        result = f(1., self.a, self.b)
        # default a is upper triangular
        expected = np.array([[13., 16., -5.],
                             [ 5.,  6., -2.]])
        assert_array_almost_equal(result, expected)

    @parametrize_blas("trmm", "sdcz")
    def test_ab_lower(self, f, dtype):
        result = f(1., self.a, self.b, lower=True)
        expected = np.array([[ 3.,  4., -1.],
                             [-1., -2.,  0.]])  # now a is lower triangular
        assert_array_almost_equal(result, expected)

    @parametrize_blas("trmm", "sdcz")
    def test_b_overwrites(self, f, dtype):
        # BLAS *trmm modifies B argument in-place.
        # Here the default is to copy, but this can be overridden
        b = self.b.astype(dtype)
        for overwr in [True, False]:
            bcopy = b.copy()
            result = f(1., self.a, bcopy, overwrite_b=overwr)
            # C-contiguous arrays are copied
            assert not bcopy.flags.f_contiguous
            assert not np.may_share_memory(bcopy, result)
            assert_equal(bcopy, b)

        bcopy = np.asfortranarray(b.copy())  # or just transpose it
        result = f(1., self.a, bcopy, overwrite_b=True)
        assert bcopy.flags.f_contiguous
        assert np.may_share_memory(bcopy, result)
        assert_array_almost_equal(bcopy, result)


@parametrize_blas("trsm", "sdcz")
def test_trsm(f, dtype):
    rng = np.random.default_rng(1234)
    tol = np.finfo(dtype).eps*1000
    func = f

    # Test protection against size mismatches
    A = rng.random((4, 5)).astype(dtype)
    B = rng.random((4, 4)).astype(dtype)
    alpha = dtype(1)
    assert_raises(Exception, func, alpha, A, B)
    assert_raises(Exception, func, alpha, A.T, B)

    n = 8
    m = 7
    alpha = dtype(-2.5)
    if dtype in COMPLEX_DTYPES:
        A = (rng.random((m, m)) + rng.random((m, m))*1j) + eye(m)
    else:
        A = rng.random((m, m)) + eye(m)
    A = A.astype(dtype)
    Au = triu(A)
    Al = tril(A)
    B1 = rng.random((m, n)).astype(dtype)
    B2 = rng.random((n, m)).astype(dtype)

    x1 = func(alpha=alpha, a=A, b=B1)
    assert_equal(B1.shape, x1.shape)
    x2 = solve(Au, alpha*B1)
    assert_allclose(x1, x2, atol=tol)

    x1 = func(alpha=alpha, a=A, b=B1, trans_a=1)
    x2 = solve(Au.T, alpha*B1)
    assert_allclose(x1, x2, atol=tol)

    x1 = func(alpha=alpha, a=A, b=B1, trans_a=2)
    x2 = solve(Au.conj().T, alpha*B1)
    assert_allclose(x1, x2, atol=tol)

    x1 = func(alpha=alpha, a=A, b=B1, diag=1)
    Au[arange(m), arange(m)] = dtype(1)
    x2 = solve(Au, alpha*B1)
    assert_allclose(x1, x2, atol=tol)

    x1 = func(alpha=alpha, a=A, b=B2, diag=1, side=1)
    x2 = solve(Au.conj().T, alpha*B2.conj().T)
    assert_allclose(x1, x2.conj().T, atol=tol)

    x1 = func(alpha=alpha, a=A, b=B2, diag=1, side=1, lower=1)
    Al[arange(m), arange(m)] = dtype(1)
    x2 = solve(Al.conj().T, alpha*B2.conj().T)
    assert_allclose(x1, x2.conj().T, atol=tol)


@pytest.mark.xfail(run=False,
                   reason="gh-16930")
def test_gh_169309():
    x = np.repeat(10, 9)
    actual = scipy.linalg.blas.dnrm2(x, 5, 3, -1)
    expected = math.sqrt(500)
    assert_allclose(actual, expected)


@pytest.mark.xfail(HAS_ILP64, reason='does not fail with ILP64 MKL')
def test_dnrm2_neg_incx():
    # check that dnrm2(..., incx < 0) raises
    # XXX: remove the test after the lowest supported BLAS implements
    # negative incx (new in LAPACK 3.10)
    x = np.repeat(10, 9)
    incx = -1
    dnrm2 = scipy.linalg.blas.get_blas_funcs('nrm2', (x,), ilp64='preferred')
    with assert_raises(FBLAS_ERROR):
        dnrm2(x, 5, 3, incx)



from types import EllipsisType

from scipy._lib._array_api import (
    array_namespace,
    Array,
    ArrayLike,
    is_lazy_array,
    xp_vector_norm,
    xp_result_type,
    xp_promote,
    is_array_api_obj,
)
import scipy._external.array_api_extra as xpx
from scipy.spatial.transform._rotation_xp import (
    as_matrix as quat_as_matrix,
    from_matrix as quat_from_matrix,
    _from_matrix_orthogonal as quat_from_matrix_orthogonal,
    from_rotvec as quat_from_rotvec,
    as_rotvec as quat_as_rotvec,
    compose_quat,
    from_quat,
    inv as quat_inv,
    mean as quat_mean,
)
from scipy._external.array_api_compat import device as xp_device
from scipy._lib._util import broadcastable


def from_matrix(matrix: Array, normalize: bool = True, copy: bool = True) -> Array:
    xp = array_namespace(matrix)
    # Shape check should be done before calling this function
    if normalize or copy:
        matrix = xp.asarray(matrix, copy=True)

    last_row_ok = xp.all(
        matrix[..., 3, :] == xp.asarray([0, 0, 0, 1.0], device=xp_device(matrix)),
        axis=-1,
    )
    lazy = is_lazy_array(matrix)
    # We delay lazy branch checks until after normalization to avoid overwriting nans
    # with the rotation matrix
    if not lazy and xp.any(~last_row_ok):
        if last_row_ok.shape == ():
            idx = ()
        else:
            idx = tuple(int(i[0]) for i in xp.nonzero(~last_row_ok))
        vals = matrix[idx + (3, ...)]
        raise ValueError(
            f"Expected last row of transformation matrix {idx} to be "
            f"exactly [0, 0, 0, 1], got {vals}"
        )

    # The quat_from_matrix() method orthogonalizes the rotation
    # component of the transformation matrix. While this does have some
    # overhead in converting a rotation matrix to a quaternion and back, it
    # allows for skipping singular value decomposition for near-orthogonal
    # matrices, which is a computationally expensive operation.
    if normalize:
        rotmat = quat_as_matrix(quat_from_matrix(matrix[..., :3, :3]))
        matrix = xpx.at(matrix)[..., :3, :3].set(rotmat)
    # Lazy branch matrix invalidation
    if lazy:
        matrix = xp.where(last_row_ok[..., None, None], matrix, xp.nan)
    return matrix


def from_rotation(quat: Array) -> Array:
    xp = array_namespace(quat)
    rotmat = quat_as_matrix(quat)
    matrix = xp.zeros(
        (*rotmat.shape[:-2], 4, 4), dtype=quat.dtype, device=xp_device(quat)
    )
    matrix = xpx.at(matrix)[..., :3, :3].set(rotmat)
    matrix = xpx.at(matrix)[..., 3, 3].set(1)
    return matrix


def from_translation(translation: Array) -> Array:
    xp = array_namespace(translation)

    if translation.shape[-1] != 3:
        raise ValueError(
            f"Expected `translation` to have shape (..., 3), got {translation.shape}."
        )
    device = xp_device(translation)
    dtype = xp_result_type(translation, force_floating=True, xp=xp)
    eye = xp.eye(4, dtype=dtype, device=device)

    matrix = xpx.atleast_nd(eye, ndim=translation.ndim + 1, xp=xp)
    matrix = xp.zeros(
        (*translation.shape[:-1], 4, 4),
        dtype=dtype,
        device=device,
    )
    matrix = xpx.at(matrix)[...].set(xp.eye(4, dtype=dtype, device=device))
    matrix = xpx.at(matrix)[..., :3, 3].set(
        xp_promote(translation, force_floating=True, xp=xp)
    )
    return matrix


def from_components(translation: Array, quat: Array) -> Array:
    return compose_transforms(from_translation(translation), from_rotation(quat))


def from_exp_coords(exp_coords: Array) -> Array:
    rot_vec = exp_coords[..., :3]
    rot_matrix = quat_as_matrix(quat_from_rotvec(rot_vec))
    translation_transform = _compute_se3_exp_translation_transform(rot_vec)
    translations = (translation_transform @ exp_coords[..., 3:, None])[..., 0]
    return _create_transformation_matrix(translations, rot_matrix)


def from_dual_quat(dual_quat: Array, *, scalar_first: bool = False) -> Array:
    xp = array_namespace(dual_quat)

    if dual_quat.shape[-1] != 8:
        raise ValueError(
            f"Expected `dual_quat` to have shape (..., 8), got {dual_quat.shape}."
        )

    real_part = dual_quat[..., :4]
    dual_part = dual_quat[..., 4:]
    if scalar_first:
        real_part = xp.roll(real_part, -1, axis=-1)
        dual_part = xp.roll(dual_part, -1, axis=-1)

    real_part, dual_part = _normalize_dual_quaternion(real_part, dual_part)

    rot_quat = from_quat(real_part)

    translation = 2.0 * compose_quat(dual_part, quat_inv(rot_quat))[..., :3]
    matrix = _create_transformation_matrix(translation, quat_as_matrix(rot_quat))

    return matrix


def as_exp_coords(matrix: Array) -> Array:
    xp = array_namespace(matrix)
    rot_vec = quat_as_rotvec(quat_from_matrix_orthogonal(matrix[..., :3, :3]))
    translation_transform = _compute_se3_log_translation_transform(rot_vec)
    translations = (translation_transform @ matrix[..., :3, 3][..., None])[..., 0]
    exp_coords = xp.concat([rot_vec, translations], axis=-1)
    return exp_coords


def as_dual_quat(matrix: Array, *, scalar_first: bool = False) -> Array:
    xp = array_namespace(matrix)
    real_parts = quat_from_matrix_orthogonal(matrix[..., :3, :3])

    pure_translation_quats = xp.empty(
        (*matrix.shape[:-2], 4), dtype=matrix.dtype, device=xp_device(matrix)
    )
    pure_translation_quats = xpx.at(pure_translation_quats)[..., :3].set(
        matrix[..., :3, 3]
    )
    pure_translation_quats = xpx.at(pure_translation_quats)[..., 3].set(0.0)

    dual_parts = 0.5 * compose_quat(pure_translation_quats, real_parts)

    if scalar_first:
        real_parts = xp.roll(real_parts, 1, axis=-1)
        dual_parts = xp.roll(dual_parts, 1, axis=-1)

    dual_quats = xp.concat([real_parts, dual_parts], axis=-1)
    return dual_quats


def compose_transforms(tf_matrix: Array, other_tf_matrix: Array) -> Array:
    if not broadcastable(tf_matrix.shape, other_tf_matrix.shape):
        len_a = tf_matrix.shape[0] if tf_matrix.ndim == 3 else 1
        len_b = other_tf_matrix.shape[0] if other_tf_matrix.ndim == 3 else 1
        raise ValueError(
            "Expected equal number of transforms in both or a "
            "single transform in either object, got "
            f"{len_a} transforms in first and {len_b}"
            "transforms in second object."
        )
    return tf_matrix @ other_tf_matrix


def inv(matrix: Array) -> Array:
    xp = array_namespace(matrix)
    r_inv = xp.matrix_transpose(matrix[..., :3, :3])
    # Matrix multiplication of r_inv and translation vector
    t_inv = -(r_inv @ matrix[..., :3, 3][..., None])[..., 0]
    matrix = xp.zeros(
        (*matrix.shape[:-2], 4, 4), dtype=matrix.dtype, device=xp_device(matrix)
    )
    matrix = xpx.at(matrix)[..., :3, :3].set(r_inv)
    matrix = xpx.at(matrix)[..., :3, 3].set(t_inv)
    matrix = xpx.at(matrix)[..., 3, 3].set(1)
    return matrix


def apply(matrix: Array, vector: Array, inverse: bool = False) -> Array:
    xp = array_namespace(matrix)
    if vector.shape[-1] != 3:
        raise ValueError(f"Expected vector to have shape (..., 3), got {vector.shape}.")
    vec = xp.empty(
        (*vector.shape[:-1], 4), dtype=vector.dtype, device=xp_device(vector)
    )
    vec = xpx.at(vec)[..., :3].set(vector)
    vec = xpx.at(vec)[..., 3].set(1)
    vec = vec[..., None]

    if inverse:
        matrix = inv(matrix)

    # We raise a ValueError manually here because letting the function run its course
    # would raise heterogeneous error types and messages for different frameworks.
    # However, the error only mimics numpy's error message and does not provide the
    # same amount of context.
    if not broadcastable(matrix.shape, vec.shape):  # type:ignore[arg-type]
        raise ValueError("operands could not be broadcast together")
    return (matrix @ vec)[..., :3, 0]


def pow(matrix: Array, n: float | Array) -> Array:
    xp = array_namespace(matrix)
    device = xp_device(matrix)
    # If n is an array, we sanitize it to a scalar and promote quat and n to
    # the same dtype.
    if is_array_api_obj(n):
        if n.shape == (1,):  # type:ignore[union-attr]
            n = n[0]  # type:ignore[index]
        elif n.ndim != 0:  # type:ignore[union-attr]
            raise ValueError("Array exponent must be a scalar")
        matrix, n = xp_promote(matrix, n, force_floating=True, xp=xp)

    # If n is a lazy array, we cannot take fast paths for special cases.
    if is_lazy_array(n):
        # Lazy execution. We compute all special cases and the general case
        result = from_exp_coords(as_exp_coords(matrix) * n)
        identity = xp.eye(4, dtype=matrix.dtype, device=device)
        result = xp.where(n == 0, identity, result)
        result = xp.where(n == -1, inv(matrix), result)
        result = xp.where(n == 1, matrix, result)
        return result
    if n == 0:
        identity = xp.eye(4, dtype=matrix.dtype, device=device)
        identity = xpx.atleast_nd(identity, ndim=matrix.ndim, xp=xp)
        return xp.tile(identity, (*matrix.shape[:-2], 1, 1))
    elif n == -1:
        return inv(matrix)
    elif n == 1:
        return matrix
    return from_exp_coords(as_exp_coords(matrix) * n)


def mean(
    matrix: Array,
    weights: ArrayLike | None = None,
    axis: None | int | tuple[int, ...] = None
) -> Array:
    xp = array_namespace(matrix)
    if matrix.shape[0] == 0:
        raise ValueError("Mean of an empty rotation set is undefined.")
    # Axis logic: For None, we reduce over all axes. For int, we only reduce over that
    # axis. For tuple, we reduce over all specified axes.
    all_axes = tuple(range(matrix.ndim - 2))
    if axis is None:
        axis = all_axes
    elif isinstance(axis, int):
        axis = (axis,)
    if not isinstance(axis, tuple):
        raise ValueError("`axis` must be None, int, or tuple of ints.")
    # Ensure all axes are within bounds
    if (axis != () and
       (min(axis) < -(matrix.ndim - 2) or max(axis) > (matrix.ndim - 3))
    ):
        raise ValueError(
            f"axis {axis} is out of bounds for transform with shape "
            f"{matrix.shape[:-2]}."
        )
    # Ensure all axes are positive and unique
    axis = tuple(sorted(set(x % (matrix.ndim - 2) for x in axis)))

    lazy = is_lazy_array(matrix)
    quats = quat_from_matrix_orthogonal(matrix[..., :3, :3])
    if weights is None:
        quats_mean = quat_mean(quats, axis=axis)
    else:
        neg_weights = weights < 0  # type:ignore[operator]
        any_neg_weights = xp.any(neg_weights)
        if not lazy and any_neg_weights:
            raise ValueError("`weights` must be non-negative.")
        if weights.shape != matrix.shape[:-2]:  # type:ignore[union-attr]
            raise ValueError(
                f"Expected `weights` to match transform shape, got shape "
                f"{weights.shape} for {matrix.shape[:-2]} transformations."  # type:ignore[union-attr]
            )
        quats_mean = quat_mean(quats, weights=weights, axis=axis)
    r_mean = quat_as_matrix(quats_mean)

    t = matrix[..., :3, 3]
    if weights is None:
        t_mean = xp.mean(t, axis=axis)
    else:
        norm = xp.sum(weights[..., None], axis=axis)  # type:ignore[index]
        wsum = xp.sum(t * weights[..., None], axis=axis)  # type:ignore[index]
        t_mean = wsum / norm

    tf = _create_transformation_matrix(t_mean, r_mean)
    if weights is not None and lazy:
        # We cannot raise on negative weights because jit code needs to be
        # non-branching. We return NaN instead
        mask = xp.where(any_neg_weights, xp.nan, 1.0)  # pyrefly:ignore[unbound-name]
        tf = mask * tf
    return tf


def setitem(
    matrix: Array,
    indexer: Array | int | tuple | slice | EllipsisType | None,
    value: Array,
) -> Array:
    xp = array_namespace(matrix)
    if isinstance(indexer, EllipsisType):
        return xpx.at(matrix)[indexer].set(value)
    if is_array_api_obj(indexer) and indexer.dtype == xp.bool:  # type:ignore[union-attr]
        return xpx.at(matrix)[indexer].set(value)  # type:ignore[index]
    return xpx.at(matrix)[indexer, ...].set(value)  # type:ignore[index]


def normalize_dual_quaternion(dual_quat: Array) -> Array:
    """Normalize dual quaternion."""
    xp = array_namespace(dual_quat)
    real, dual = _normalize_dual_quaternion(dual_quat[..., :4], dual_quat[..., 4:])
    return xp.concat((real, dual), axis=-1)


def _create_transformation_matrix(
    translations: Array, rotation_matrices: Array
) -> Array:
    if not translations.shape[:-1] == rotation_matrices.shape[:-2]:
        raise ValueError(
            "The number of rotation matrices and translations must be the same."
        )
    xp = array_namespace(translations)
    matrix = xp.empty(
        (*translations.shape[:-1], 4, 4),
        dtype=translations.dtype,
        device=xp_device(translations),
    )
    matrix = xpx.at(matrix)[..., :3, :3].set(rotation_matrices)
    matrix = xpx.at(matrix)[..., :3, 3].set(translations)
    matrix = xpx.at(matrix)[..., 3, :3].set(0)
    matrix = xpx.at(matrix)[..., 3, 3].set(1)
    return matrix


def _compute_se3_exp_translation_transform(rot_vec: Array) -> Array:
    """Compute the transformation matrix from the se3 translation part to SE3
    translation.

    The transformation matrix depends on the rotation vector.
    """
    xp = array_namespace(rot_vec)
    device = xp_device(rot_vec)
    dtype = rot_vec.dtype
    angle = xp_vector_norm(rot_vec, axis=-1, keepdims=True, xp=xp)
    small_scale = angle < 1e-3

    k1_small = 0.5 - angle**2 / 24 + angle**4 / 720
    # Avoid division by zero for non-branching computations. The value will get
    # discarded in the xp.where selection.
    safe_angle = angle + xp.asarray(small_scale, dtype=dtype, device=device)
    k1 = (1.0 - xp.cos(angle)) / safe_angle**2
    k1 = xp.where(small_scale, k1_small, k1)

    k2_small = 1 / 6 - angle**2 / 120 + angle**4 / 5040
    # Again, avoid division by zero by adding one to all near-zero angles.
    safe_angle = angle + xp.asarray(small_scale, dtype=dtype, device=device)
    k2 = (angle - xp.sin(angle)) / safe_angle**3
    k2 = xp.where(small_scale, k2_small, k2)

    s = _create_skew_matrix(rot_vec)
    eye = xp.eye(3, dtype=dtype, device=device)

    return eye + k1[..., None] * s + k2[..., None] * s @ s


def _compute_se3_log_translation_transform(rot_vec: Array) -> Array:
    """Compute the transformation matrix from SE3 translation to the se3
    translation part.

    It is the inverse of `_compute_se3_exp_translation_transform` in a closed
    analytical form.
    """
    xp = array_namespace(rot_vec)
    dtype = rot_vec.dtype
    device = xp_device(rot_vec)
    angle = xp_vector_norm(rot_vec, axis=-1, keepdims=True, xp=xp)
    mask = angle < 1e-3

    k_small = 1 / 12 + angle**2 / 720 + angle**4 / 30240
    safe_angle = angle + xp.asarray(mask, dtype=dtype, device=device)
    k = (1 - 0.5 * angle / xp.tan(0.5 * safe_angle)) / safe_angle**2
    k = xp.where(mask, k_small, k)

    s = _create_skew_matrix(rot_vec)

    return xp.eye(3, dtype=dtype, device=device) - 0.5 * s + k[..., None] * s @ s


def _create_skew_matrix(vec: Array) -> Array:
    """Create skew-symmetric (aka cross-product) matrix for stack of vectors."""
    xp = array_namespace(vec)
    result = xp.zeros((*vec.shape[:-1], 3, 3), dtype=vec.dtype, device=xp_device(vec))
    result = xpx.at(result)[..., 0, 1].set(-vec[..., 2])
    result = xpx.at(result)[..., 0, 2].set(vec[..., 1])
    result = xpx.at(result)[..., 1, 0].set(vec[..., 2])
    result = xpx.at(result)[..., 1, 2].set(-vec[..., 0])
    result = xpx.at(result)[..., 2, 0].set(-vec[..., 1])
    result = xpx.at(result)[..., 2, 1].set(vec[..., 0])
    return result


def _normalize_dual_quaternion(
    real_part: Array, dual_part: Array
) -> tuple[Array, Array]:
    """Ensure that the dual quaternion has unit norm.

    The norm is a dual number and must be 1 + 0 * epsilon, which means that
    the real quaternion must have unit norm and the dual quaternion must be
    orthogonal to the real quaternion.
    """
    xp = array_namespace(real_part)

    real_norm = xp_vector_norm(real_part, axis=-1, keepdims=True, xp=xp)

    # special case: real quaternion is 0, we set it to identity
    zero_real_mask = real_norm == 0.0
    unit_quat = xp.asarray(
        [0.0, 0.0, 0.0, 1.0], dtype=real_part.dtype, device=xp_device(real_part)
    )
    real_part = xp.where(zero_real_mask, unit_quat, real_part)
    real_norm = xp.where(zero_real_mask, 1.0, real_norm)

    # 1. ensure unit real quaternion
    real_part = real_part / real_norm
    dual_part = dual_part / real_norm

    # 2. ensure orthogonality of real and dual quaternion
    dual_part -= xp.sum(real_part * dual_part, axis=-1, keepdims=True) * real_part

    return real_part, dual_part



# This file is not meant for public use and will be removed in SciPy v2.0.0.
# Use the `scipy.odr` namespace for importing the functions
# included below.


__all__ = [  # noqa: F822
    'odr', 'OdrWarning', 'OdrError', 'OdrStop',
    'Data', 'RealData', 'Model', 'Output', 'ODR',
    'odr_error', 'odr_stop'
]


def __dir__():
    return __all__


def __getattr__(name):
    msg = ("`scipy.odr` is deprecated as of version 1.17.0 and will be removed in "
           "SciPy 1.19.0. Please use `https://pypi.org/project/odrpack/` instead.")
    if name not in __all__:
        raise AttributeError(
            f"`scipy.odr.odrpack` has no attribute {name}. In addition, {msg}")

    import warnings
    from . import _odrpack
    warnings.warn(msg, category=DeprecationWarning, stacklevel=2)

    return getattr(_odrpack, name)



import numpy as np
from functools import partial

from scipy import stats
from scipy.stats._axis_nan_policy import _broadcast_arrays
from scipy._external import array_api_extra as xpx
from scipy._lib._array_api import (xp_capabilities, array_namespace, xp_ravel,
                                   xp_result_type)


def _bws_input_validation(x, y, alternative, axis, method):
    ''' Input validation and standardization for bws test'''
    xp = array_namespace(x, y)
    x, y = xp.asarray(x), xp.asarray(y)
    x, y = xpx.atleast_nd(x, ndim=1), xpx.atleast_nd(y, ndim=1)

    axis_none = axis is None
    ndim = max(x.ndim, y.ndim)
    if axis_none:
        x = xp_ravel(x, xp=xp)
        y = xp_ravel(y, xp=xp)
        axis = 0
    elif np.iterable(axis) or int(axis) != axis:
        message = "`axis` must be an integer or None."
        raise ValueError(message)
    elif (axis >= ndim) or (axis < -ndim):
        message = "`axis` is not compatible with the shapes of the inputs."
        raise ValueError(message)
    axis = int(axis)

    x, y = _broadcast_arrays((x, y), axis=axis, xp=xp)
    nx, ny = x.shape[-1], y.shape[-1]
    if nx < 2 or ny < 2:
        raise ValueError('`x` and `y` must contain at least two observations each.')

    x, y = xp.moveaxis(x, axis, -1), xp.moveaxis(y, axis, -1)
    z = stats.rankdata(xp.concat((x, y), axis=-1), axis=-1)
    z = xp.astype(z, xp_result_type(x, y, force_floating=True, xp=xp))
    x, y = z[..., :x.shape[-1]], z[..., x.shape[-1]:]
    x, y = xp.moveaxis(x, -1, axis), xp.moveaxis(y, -1, axis)

    alternatives = {'two-sided', 'less', 'greater'}
    alternative = alternative.lower()
    if alternative not in alternatives:
        raise ValueError(f'`alternative` must be one of {alternatives}.')

    method = stats.PermutationMethod() if method is None else method
    if not isinstance(method, stats.PermutationMethod):
        raise ValueError('`method` must be an instance of '
                         '`scipy.stats.PermutationMethod`')

    return x, y, alternative, axis, method, xp


def _bws_statistic(x, y, alternative, axis, xp):
    '''Compute the BWS test statistic for two independent samples'''
    # Public function currently does not accept `axis`, but `permutation_test`
    # uses `axis` to make vectorized call.

    Ri, Hj = xp.sort(x, axis=axis), xp.sort(y, axis=axis)
    n, m = Ri.shape[axis], Hj.shape[axis]
    i, j = xp.arange(1, n+1, dtype=Ri.dtype), xp.arange(1, m+1, dtype=Hj.dtype)

    Bx_num = Ri - (m + n)/n * i
    By_num = Hj - (m + n)/m * j

    if alternative == 'two-sided':
        Bx_num = xpx.at(Bx_num)[...].multiply(Bx_num)
        By_num = xpx.at(By_num)[...].multiply(By_num)
    else:
        Bx_num = xpx.at(Bx_num)[...].multiply(xp.abs(Bx_num))
        By_num = xpx.at(By_num)[...].multiply(xp.abs(By_num))

    Bx_den = i/(n+1) * (1 - i/(n+1)) * m*(m+n)/n
    By_den = j/(m+1) * (1 - j/(m+1)) * n*(m+n)/m

    Bx = 1/n * xp.sum(Bx_num/Bx_den, axis=axis)
    By = 1/m * xp.sum(By_num/By_den, axis=axis)

    B = (Bx + By) / 2 if alternative == 'two-sided' else (Bx - By) / 2

    return B


@xp_capabilities(skip_backends=[('dask.array', 'no rankdata')])
def bws_test(x, y, *, alternative="two-sided", axis=0, method=None):
    r'''Perform the Baumgartner-Weiss-Schindler test on two independent samples.

    The Baumgartner-Weiss-Schindler (BWS) test is a nonparametric test of
    the null hypothesis that the distribution underlying sample `x`
    is the same as the distribution underlying sample `y`. Unlike
    the Kolmogorov-Smirnov, Wilcoxon, and Cramer-Von Mises tests,
    the BWS test weights the integral by the variance of the difference
    in cumulative distribution functions (CDFs), emphasizing the tails of the
    distributions, which increases the power of the test in many applications.

    Parameters
    ----------
    x, y : array-like
        Arrays of samples. Shapes must be broadcastable except along `axis`.
    alternative : {'two-sided', 'less', 'greater'}, optional
        Defines the alternative hypothesis. Default is 'two-sided'.
        Let *F(u)* and *G(u)* be the cumulative distribution functions of the
        distributions underlying `x` and `y`, respectively. Then the following
        alternative hypotheses are available:

        * 'two-sided': the distributions are not equal, i.e. *F(u) ≠ G(u)* for
          at least one *u*.
        * 'less': the distribution underlying `x` is stochastically less than
          the distribution underlying `y`, i.e. *F(u) >= G(u)* for all *u*.
        * 'greater': the distribution underlying `x` is stochastically greater
          than the distribution underlying `y`, i.e. *F(u) <= G(u)* for all
          *u*.

        Under a more restrictive set of assumptions, the alternative hypotheses
        can be expressed in terms of the locations of the distributions;
        see [2] section 5.1.
    axis : int or None, default: 0
        Axis along which the quantiles are computed.
        ``None`` ravels both `x` and `y` before performing the calculation.
    method : PermutationMethod, optional
        Configures the method used to compute the p-value. The default is
        the default `PermutationMethod` object.

    Returns
    -------
    res : PermutationTestResult
        An object with attributes:

        statistic : float
            The observed test statistic of the data.
        pvalue : float
            The p-value for the given alternative.
        null_distribution : ndarray
            The values of the test statistic generated under the null hypothesis.

    See Also
    --------
    scipy.stats.wilcoxon, scipy.stats.mannwhitneyu, scipy.stats.ttest_ind

    Notes
    -----
    When ``alternative=='two-sided'``, the statistic is defined by the
    equations given in [1]_ Section 2. This statistic is not appropriate for
    one-sided alternatives; in that case, the statistic is the *negative* of
    that given by the equations in [1]_ Section 2. Consequently, when the
    distribution of the first sample is stochastically greater than that of the
    second sample, the statistic will tend to be positive.

    References
    ----------
    .. [1] Neuhäuser, M. (2005). Exact Tests Based on the
           Baumgartner-Weiss-Schindler Statistic: A Survey. Statistical Papers,
           46(1), 1-29.
    .. [2] Fay, M. P., & Proschan, M. A. (2010). Wilcoxon-Mann-Whitney or t-test?
           On assumptions for hypothesis tests and multiple interpretations of
           decision rules. Statistics surveys, 4, 1.

    Examples
    --------
    We follow the example of table 3 in [1]_: Fourteen children were divided
    randomly into two groups. Their ranks at performing a specific tests are
    as follows.

    >>> import numpy as np
    >>> x = [1, 2, 3, 4, 6, 7, 8]
    >>> y = [5, 9, 10, 11, 12, 13, 14]

    We use the BWS test to assess whether there is a statistically significant
    difference between the two groups.
    The null hypothesis is that there is no difference in the distributions of
    performance between the two groups. We decide that a significance level of
    1% is required to reject the null hypothesis in favor of the alternative
    that the distributions are different.
    Since the number of samples is very small, we can compare the observed test
    statistic against the *exact* distribution of the test statistic under the
    null hypothesis.

    >>> from scipy.stats import bws_test
    >>> res = bws_test(x, y)
    >>> print(res.statistic)
    5.132167152575315

    This agrees with :math:`B = 5.132` reported in [1]_. The *p*-value produced
    by `bws_test` also agrees with :math:`p = 0.0029` reported in [1]_.

    >>> print(res.pvalue)
    0.002913752913752914

    Because the p-value is below our threshold of 1%, we take this as evidence
    against the null hypothesis in favor of the alternative that there is a
    difference in performance between the two groups.
    '''

    x, y, alternative, axis, method, xp = _bws_input_validation(x, y, alternative,
                                                                axis, method)
    bws_statistic = partial(_bws_statistic, alternative=alternative, xp=xp)

    permutation_alternative = 'less' if alternative == 'less' else 'greater'
    res = stats.permutation_test((x, y), bws_statistic,
                                 alternative=permutation_alternative, axis=axis,
                                 **method._asdict())

    return res



from typing import Final

from ..._internal import clone_module

__all__ = clone_module("dask.array", globals())

# These imports may overwrite names from the import * above.
from . import _aliases
from ._aliases import *  # type: ignore[assignment] # noqa: F403
from ._info import __array_namespace_info__  # noqa: F401

__array_api_version__: Final = "2025.12"
del Final

# See the comment in the numpy __init__.py
__import__(__package__ + '.linalg')
__import__(__package__ + '.fft')

__all__ = sorted(
    set(__all__)
    | set(_aliases.__all__)
    | {"__array_api_version__", "__array_namespace_info__", "linalg", "fft"}
)

def __dir__() -> list[str]:
    return __all__



# Last Change: Sat Mar 21 02:00 PM 2009 J

# Copyright (c) 2001, 2002 Enthought, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the Enthought nor the names of its contributors
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

"""Some more special functions which may be useful for multivariate statistical
analysis."""

import numpy as np

import scipy.special._gufuncs as _gufuncs
from scipy.special import gammaln as loggam
from scipy.special._ufunc_tools import _with_cache_optimization


__all__ = ['multigammaln']


def multigammaln(a, d):
    r"""Returns the log of multivariate gamma, also sometimes called the
    generalized gamma.

    Parameters
    ----------
    a : ndarray
        The multivariate gamma is computed for each item of `a`.
    d : int
        The dimension of the space of integration.

    Returns
    -------
    res : ndarray
        The values of the log multivariate gamma at the given points `a`.

    Notes
    -----
    The formal definition of the multivariate gamma of dimension d for a real
    `a` is

    .. math::

        \Gamma_d(a) = \int_{A>0} e^{-tr(A)} |A|^{a - (d+1)/2} dA

    with the condition :math:`a > (d-1)/2`, and :math:`A > 0` being the set of
    all the positive definite matrices of dimension `d`.  Note that `a` is a
    scalar: the integrand only is multivariate, the argument is not (the
    function is defined over a subset of the real set).

    This can be proven to be equal to the much friendlier equation

    .. math::

        \Gamma_d(a) = \pi^{d(d-1)/4} \prod_{i=1}^{d} \Gamma(a - (i-1)/2).

    References
    ----------
    R. J. Muirhead, Aspects of multivariate statistical theory (Wiley Series in
    probability and mathematical statistics).

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.special import multigammaln, gammaln
    >>> a = 23.5
    >>> d = 10
    >>> multigammaln(a, d)
    454.1488605074416

    Verify that the result agrees with the logarithm of the equation
    shown above:

    >>> d*(d-1)/4*np.log(np.pi) + gammaln(a - 0.5*np.arange(0, d)).sum()
    454.1488605074416
    """
    a = np.asarray(a)
    # Support for 0d arrays is needed for array_api_strict and dask.
    d = np.asarray(d)[()]
    if not np.isscalar(d) or (np.floor(d) != d):
        raise ValueError("d should be a positive integer (dimension)")
    if np.any(a <= 0.5 * (d - 1)):
        raise ValueError(f"condition a ({a}) > 0.5 * (d-1) ({0.5 * (d-1)}) not met")

    res = (d * (d-1) * 0.25) * np.log(np.pi)
    res += np.sum(loggam([(a - (j - 1.)/2) for j in range(1, d+1)]), axis=0)
    return res


_poisson_binom_pmf_doc = (
    """Returns pmf of Poisson Binomial distribution.

    Parameters
    ----------
    k : array
        Number of successes at which to evaluate pmf.

    p : array
        Success probabilities of independent Bernoulli trials.

    Notes
    -----
    This is equivalent to a gufunc with signature ``()(i)->()``.
    The last dimension of `p` contains success probabilities and
    the preceding dimensions are batch dimensions. The batch
    dimensions are broadcast against ``k``.

    The output will be C contiguous regardless of the contiguity of
    `k` and `p`.

    """
)


_poisson_binom_pmf = _with_cache_optimization(
    name="_poisson_binom_pmf",
    arg_names=["k", "p"],
    docstring=_poisson_binom_pmf_doc,
    ufunc=_gufuncs._poisson_binom_pmf,
    cache_arg_indices=[1],
)


_poisson_binom_cdf_doc = (
    """Returns cdf of Poisson Binomial distribution.

    Parameters
    ----------
    k : array
        Number of successes at which to evaluate cdf.

    p : array
        Success probabilities of independent Bernoulli trials.

    Notes
    -----
    This is equivalent to a gufunc with signature ``()(i)->()``.
    The last dimension of `p` contains success probabilities and
    the preceding dimensions are batch dimensions. The batch
    dimensions are broadcast against ``k``.

    The output will be C contiguous regardless of the contiguity of
    `k` and `p`.

    """
)

_poisson_binom_cdf = _with_cache_optimization(
    name="_poisson_binom_cdf",
    arg_names=["k", "p"],
    docstring=_poisson_binom_cdf_doc,
    ufunc=_gufuncs._poisson_binom_cdf,
    cache_arg_indices=[1],
)



import pytest

from scipy._lib._array_api import np_compat as np
from scipy._lib._array_api import (
    array_namespace, make_xp_pytest_param, xp_assert_close, xp_capabilities
)

local_capabilities_table = {}  # type:ignore[var-annotated]

# B is a child of A which inherits the method g which is array-agnostic
# so long as the method f is supported. A.f does not support the JAX jit but
# B.f does support the JAX jit. Test that this inheritence does not
# cause problems when testing with JAX jit.

@xp_capabilities(
    capabilities_table=local_capabilities_table,
    cpu_only=True,
    jax_jit=False,
    skip_backends=[
        ("dask.array", ""),
    ],
)
class A:
    def __init__(self, x):
        xp = array_namespace(x)
        self._xp = xp
        self.x = np.asarray(x)

    def f(self, y):
        y = np.asarray(y)
        return self._xp.asarray(np.matmul(self.x, y))

    def g(self, y, z):
        return self.f(y) + self.f(z)


@xp_capabilities(
    capabilities_table=local_capabilities_table,
    cpu_only=True,
    skip_backends=[("dask.array", "")],
)
class B(A):
    def __init__(self, x):
        xp = array_namespace(x)
        self._xp = xp
        self.x = xp.asarray(x)

    def f(self, y):
        return self._xp.matmul(self.x, y)


@pytest.mark.parametrize(
    "cls",
    [
        make_xp_pytest_param((A, "g"), capabilities_table=local_capabilities_table),
        make_xp_pytest_param((B, "g"), capabilities_table=local_capabilities_table),
    ],
)
def test_no_spooky_action_at_a_distance(cls, xp):
    # test that application of lazy_xp_function to a method which is inherited
    # from a parent class do not propagate to the parent class. Calls to A.g
    # will raise here if A.g was accidentally jitted.
    x = xp.asarray([1.1, 2.2, 3.3])
    y = xp.asarray([1.0, 2.0, 3.0])
    z = xp.asarray([3.0, 4.0, 5.0])
    foo = cls(x)
    observed = foo.g(y, z)
    expected = xp.asarray(44.0)[()]
    xp_assert_close(observed, expected)



"""Array API backend for the `Rotation` class.

This module provides generic, functional implementations of the `Rotation` class methods
that work with any Array API-compatible backend.
"""

# Parts of the implementation are adapted from the cython backend and
# https://github.com/jax-ml/jax/blob/d695aa4c63ffcebefce52794427c46bad576680c/jax/_src/scipy/spatial/transform.py.
import re
import warnings
from types import EllipsisType

import numpy as np
from scipy._lib._array_api import (
    array_namespace,
    Array,
    ArrayLike,
    is_lazy_array,
    xp_vector_norm,
    xp_result_type,
    xp_promote,
    is_jax,
)
from scipy._lib._util import broadcastable
from scipy._external.array_api_compat import device as xp_device
from scipy._external.array_api_compat import is_array_api_obj
import scipy._external.array_api_extra as xpx

# mypy: disable-error-code=index
# mypy: disable-error-code=operator
# mypy: disable-error-code=union-attr

def from_quat(
    quat: Array,
    normalize: bool = True,
    copy: bool = True,
    *,
    scalar_first: bool = False,
) -> Array:
    xp = array_namespace(quat)
    if scalar_first:
        quat = xp.roll(quat, -1, axis=-1)
    # Normalize will always copy, so we avoid the extra copy if we normalize
    if copy and not normalize:
        quat = xp.asarray(quat, copy=True)
    if normalize:
        quat = _normalize_quaternion(quat)
    return quat


def from_matrix(matrix: Array, assume_valid: bool = False) -> Array:
    xp = array_namespace(matrix)
    device = xp_device(matrix)

    if not assume_valid:
        mask = xp.linalg.det(matrix) <= 0
        lazy = is_lazy_array(mask)
        # Only non-lazy backends raise an error for non-positive determinants.
        if not lazy and xp.any(mask):
            ind = int(xp.nonzero(xpx.atleast_nd(mask, ndim=1, xp=xp))[0][0])
            raise ValueError(
                "Non-positive determinant (left-handed or null coordinate frame) in "
                f"rotation matrix {ind}: {matrix[ind, ...]}."
            )
        elif lazy:
            matrix = xp.where(mask[..., None, None], xp.nan, matrix)

        gramians = matrix @ xp.matrix_transpose(matrix)
        eye = xp.eye(3, dtype=matrix.dtype, device=device)
        is_orthogonal = xp.all(
            xpx.isclose(gramians, eye, atol=1e-12, xp=xp), axis=(-2, -1)
        )

        if lazy:
            # Lazy backends do not support non-concrete boolean indexing or any form of
            # computation without statically known shapes, so we always compute SVD and
            # use xp.where to select the result.
            U, _, Vt = xp.linalg.svd(matrix, full_matrices=False)
            matrix = xp.where(is_orthogonal[..., None, None], matrix, U @ Vt)
        elif not xp.all(is_orthogonal):
            # For eager frameworks, only compute SVD if needed.
            is_not_orthogonal = ~is_orthogonal
            U, _, Vt = xp.linalg.svd(matrix[is_not_orthogonal], full_matrices=False)
            matrix = xpx.at(matrix)[is_not_orthogonal].set(U @ Vt)

    return _from_matrix_orthogonal(matrix)


def _from_matrix_orthogonal(matrix: Array) -> Array:
    """Convert known orthogonal rotation matrix to quaternion"""
    xp = array_namespace(matrix)
    device = xp_device(matrix)

    matrix_trace = matrix[..., 0, 0] + matrix[..., 1, 1] + matrix[..., 2, 2]
    decision = xp.stack(
        [matrix[..., 0, 0], matrix[..., 1, 1], matrix[..., 2, 2], matrix_trace],
        axis=-1,
    )
    choice = xp.argmax(decision, axis=-1, keepdims=True)
    quat = xp.empty((*matrix.shape[:-2], 4), dtype=matrix.dtype, device=device)

    # The Array API does not support mixing integer indexing with ellipsis, so we cannot
    # follow the same pattern as the cython backend. Instead, we compute each case
    # explicitly and assemble the final result with `xp.where`.
    # TODO: Revisit this implementation if the array API supports mixed integer and
    # ellipsis indexing.

    # Case 0
    quat_0 = xp.stack(
        [
            1 - matrix_trace[...] + 2 * matrix[..., 0, 0],
            matrix[..., 1, 0] + matrix[..., 0, 1],
            matrix[..., 2, 0] + matrix[..., 0, 2],
            matrix[..., 2, 1] - matrix[..., 1, 2],
        ],
        axis=-1,
    )
    quat = xp.where(choice == 0, quat_0, quat)

    # Case 1
    quat_1 = xp.stack(
        [
            matrix[..., 1, 0] + matrix[..., 0, 1],
            1 - matrix_trace[...] + 2 * matrix[..., 1, 1],
            matrix[..., 2, 1] + matrix[..., 1, 2],
            matrix[..., 0, 2] - matrix[..., 2, 0],
        ],
        axis=-1,
    )
    quat = xp.where(choice == 1, quat_1, quat)

    # Case 2
    quat_2 = xp.stack(
        [
            matrix[..., 2, 0] + matrix[..., 0, 2],
            matrix[..., 2, 1] + matrix[..., 1, 2],
            1 - matrix_trace[...] + 2 * matrix[..., 2, 2],
            matrix[..., 1, 0] - matrix[..., 0, 1],
        ],
        axis=-1,
    )
    quat = xp.where(choice == 2, quat_2, quat)

    # Case 3
    quat_3 = xp.stack(
        [
            matrix[..., 2, 1] - matrix[..., 1, 2],
            matrix[..., 0, 2] - matrix[..., 2, 0],
            matrix[..., 1, 0] - matrix[..., 0, 1],
            1 + matrix_trace[...],
        ],
        axis=-1,
    )
    quat = xp.where(choice == 3, quat_3, quat)

    return _normalize_quaternion(quat)


def from_rotvec(rotvec: Array, degrees: bool = False) -> Array:
    xp = array_namespace(rotvec)
    if rotvec.shape[-1] != 3:
        raise ValueError(
            f"Expected `rot_vec` to have shape (..., 3), got {rotvec.shape}"
        )
    rotvec = _deg2rad(rotvec) if degrees else rotvec

    angle = xp_vector_norm(rotvec, axis=-1, keepdims=True, xp=xp)
    small_angle = angle <= 1e-3
    angle2 = angle**2
    small_scale = 0.5 - angle2 / 48 + angle2**2 / 3840
    # We need to handle the case where angle is 0 to avoid division by zero. We use the
    # value of the Taylor series approximation, but non-branching operations require
    # that we still divide by the angle. Since we do not use the result where the angle
    # is close to 0, this is safe.
    div_angle = angle + xp.asarray(small_angle, dtype=angle.dtype)
    large_scale = xp.sin(angle / 2) / div_angle
    scale = xp.where(small_angle, small_scale, large_scale)
    quat = xp.concat([rotvec * scale, xp.cos(angle / 2)], axis=-1)
    return quat


def from_mrp(mrp: Array) -> Array:
    xp = array_namespace(mrp)
    if mrp.shape[-1] != 3:
        raise ValueError(f"Expected `mrp` to have shape (..., 3), got {mrp.shape}")
    mrp2_plus_1 = xp.linalg.vecdot(mrp, mrp, axis=-1)[..., None] + 1
    q_no_norm = xp.concat([2 * mrp[..., :3], (2 - mrp2_plus_1)], axis=-1)
    quat = q_no_norm / mrp2_plus_1
    return quat


def from_euler(seq: str, angles: Array, degrees: bool = False) -> Array:
    xp = array_namespace(angles)
    num_axes = len(seq)
    if num_axes < 1 or num_axes > 3:
        raise ValueError(
            "Expected axis specification to be a non-empty "
            f"string of up to 3 characters, got {seq}"
        )

    intrinsic = re.match(r"^[XYZ]{1,3}$", seq) is not None
    extrinsic = re.match(r"^[xyz]{1,3}$", seq) is not None
    if not (intrinsic or extrinsic):
        raise ValueError(
            "Expected axes from `seq` to be from ['x', 'y', "
            f"'z'] or ['X', 'Y', 'Z'], got {seq}"
        )

    if any(seq[i] == seq[i + 1] for i in range(num_axes - 1)):
        raise ValueError(f"Expected consecutive axes to be different, got {seq}")

    if degrees:
        angles = _deg2rad(angles)

    angles = xpx.atleast_nd(angles, ndim=1, xp=xp)

    if angles.shape[-1] != num_axes:
        raise ValueError(
            "Expected last dimension of `angles` to match number of sequence axes "
            f"specified, got {angles.shape[-1]}."
        )
    axes = [_elementary_basis_index(x) for x in seq.lower()]
    q = _elementary_quat_compose(axes, angles, intrinsic)
    return q


def from_davenport(
    axes: Array, order: str, angles: Array, degrees: bool = False
) -> Array:
    xp = array_namespace(axes)
    device = xp_device(axes)
    if order in ["e", "extrinsic"]:  # Must be static, cannot be jitted
        extrinsic = True
    elif order in ["i", "intrinsic"]:
        extrinsic = False
    else:
        raise ValueError(
            "order should be 'e'/'extrinsic' for extrinsic sequences or 'i'/"
            f"'intrinsic' for intrinsic sequences, got {order}"
        )

    if axes.shape[-1] != 3:
        raise ValueError("Axes must be vectors of length 3.")

    axes = xpx.atleast_nd(axes, ndim=2, xp=xp)
    angles = xpx.atleast_nd(angles, ndim=1, xp=xp) 
    num_axes = axes.shape[-2]
    if num_axes < 1 or num_axes > 3:
        raise ValueError(f"Expected up to 3 axes, got {num_axes}")

    axes = axes / xp_vector_norm(axes, axis=-1, keepdims=True, xp=xp)

    # Check if axes are orthogonal. Shape checks also work for lazy backends.
    axes_not_orthogonal = xp.zeros(axes.shape[:-2], dtype=xp.bool, device=device)
    if num_axes > 1:
        # Cannot be True yet, so we do not need to use xp.logical_or
        axes_not_orthogonal = axes_not_orthogonal | (
            xp.abs(xp.vecdot(axes[..., 0, :], axes[..., 1, :])) > 1e-7
        )
    if num_axes > 2:
        axes_not_orthogonal = axes_not_orthogonal | (
            xp.abs(xp.vecdot(axes[..., 1, :], axes[..., 2, :])) > 1e-7
        )
    if not is_lazy_array(axes_not_orthogonal) and xp.any(axes_not_orthogonal):
        raise ValueError("Consecutive axes must be orthogonal.")
    else:
        axes = xp.where(axes_not_orthogonal[..., None, None], xp.nan, axes)

    if degrees:
        angles = _deg2rad(angles)

    if (
        not broadcastable(axes.shape[:-1], angles.shape)
        or axes.shape[-2] != angles.shape[-1]
    ):
        raise ValueError(
            f"Expected `angles` to match number of axes, got {angles.shape} angles "
            f"and {axes.shape} axes."
        )

    q_shape = angles.shape[:-1] + (4,)
    q = xp.zeros(q_shape, dtype=angles.dtype, device=xp_device(angles))
    q = xpx.at(q)[..., 3].set(1)

    for i in range(num_axes):
        qi = from_rotvec(angles[..., i, None] * axes[..., i, :])
        q = compose_quat(qi, q) if extrinsic else compose_quat(q, qi)
    return q


def as_quat(
    quat: Array, canonical: bool = False, *, scalar_first: bool = False
) -> Array:
    xp = array_namespace(quat)
    if canonical:
        quat = _quat_canonical(quat)
    if scalar_first:
        quat = xp.roll(quat, 1, axis=-1)
    return quat


def as_matrix(quat: Array) -> Array:
    xp = array_namespace(quat)
    x = quat[..., 0]
    y = quat[..., 1]
    z = quat[..., 2]
    w = quat[..., 3]

    x2 = x * x
    y2 = y * y
    z2 = z * z
    w2 = w * w

    xy = x * y
    zw = z * w
    xz = x * z
    yw = y * w
    yz = y * z
    xw = x * w

    matrix_elements = [
        x2 - y2 - z2 + w2,
        2 * (xy - zw),
        2 * (xz + yw),
        2 * (xy + zw),
        -x2 + y2 - z2 + w2,
        2 * (yz - xw),
        2 * (xz - yw),
        2 * (yz + xw),
        -x2 - y2 + z2 + w2,
    ]
    matrix = xp.reshape(xp.stack(matrix_elements, axis=-1), (*quat.shape[:-1], 3, 3))
    return matrix


def as_rotvec(quat: Array, degrees: bool = False) -> Array:
    xp = array_namespace(quat)
    quat = _quat_canonical(quat)
    ax_norm = xp_vector_norm(quat[..., :3], axis=-1, keepdims=True, xp=xp)
    angle = 2 * xp.atan2(ax_norm, quat[..., 3][..., None])
    small_angle = angle <= 1e-3
    angle2 = angle**2
    small_scale = 2 + angle2 / 12 + 7 * angle2**2 / 2880
    # We need to handle the case where sin(angle/2) is 0 to avoid division by zero. We
    # use the value of the Taylor series approximation, but non-branching operations
    # require that we still divide by the sin. Since we do not use the result where the
    # angle is close to 0, adding one to the sin where we discard the result is safe.
    div_sin = xp.sin(angle / 2.0) + xp.asarray(small_angle, dtype=angle.dtype)
    large_scale = angle / div_sin
    scale = xp.where(small_angle, small_scale, large_scale)
    if degrees:
        scale = _rad2deg(scale)
    rotvec = scale * quat[..., :3]
    return rotvec


def as_mrp(quat: Array) -> Array:
    xp = array_namespace(quat)
    one = xp.asarray(1.0, device=xp_device(quat), dtype=quat.dtype)
    sign = xp.where(quat[..., 3, None] < 0, -1, one)
    denominator = 1.0 + sign * quat[..., 3, None]
    return sign * quat[..., :3] / denominator


def as_euler(
    quat: Array, seq: str, degrees: bool = False, *, suppress_warnings: bool = False
) -> Array:
    xp = array_namespace(quat)

    # Sanitize the sequence
    if len(seq) != 3:
        raise ValueError(f"Expected 3 axes, got {seq}.")
    intrinsic = re.match(r"^[XYZ]{1,3}$", seq) is not None
    extrinsic = re.match(r"^[xyz]{1,3}$", seq) is not None
    if not (intrinsic or extrinsic):
        raise ValueError(
            "Expected axes from `seq` to be from ['x', 'y', 'z'] or ['X', 'Y', 'Z'], "
            f"got {seq}"
        )
    if any(seq[i] == seq[i + 1] for i in range(2)):
        raise ValueError(f"Expected consecutive axes to be different, got {seq}")

    device = xp_device(quat)
    axes = [_elementary_basis_index(x) for x in seq.lower()]
    axes = axes if extrinsic else axes[::-1]
    i, j, k = axes
    symmetric = i == k
    k = 3 - i - j if symmetric else k

    mask = xp.asarray(symmetric, device=device)
    sign = xp.asarray((i - j) * (j - k) * (k - i) // 2, dtype=quat.dtype, device=device)
    # Permute quaternion elements
    a = xp.where(mask, quat[..., 3], quat[..., 3] - quat[..., j])
    b = xp.where(mask, quat[..., i], quat[..., i] + quat[..., k] * sign)
    c = xp.where(mask, quat[..., j], quat[..., j] + quat[..., 3])
    d = xp.where(mask, quat[..., k] * sign, quat[..., k] * sign - quat[..., i])

    angles = _get_angles(
        extrinsic, symmetric, sign, xp.pi / 2, a, b, c, d, suppress_warnings
    )
    return _rad2deg(angles) if degrees else angles


def as_davenport(
    quat: Array,
    axes: Array,
    order: str,
    degrees: bool = False,
    *,
    suppress_warnings: bool = False,
) -> Array:
    xp = array_namespace(quat)

    # Check argument validity
    if order in ["e", "extrinsic"]:
        extrinsic = True
    elif order in ["i", "intrinsic"]:
        extrinsic = False
    else:
        raise ValueError(
            "order should be 'e'/'extrinsic' for extrinsic sequences or 'i'/'intrinsic'"
            f" for intrinsic sequences, got {order}"
        )
    if axes.shape[-2] != 3:
        raise ValueError(f"Expected 3 axes, got {axes.shape}.")
    if axes.shape[-1] != 3:
        raise ValueError("Axes must be vectors of length 3.")
    if not broadcastable(axes.shape[:-2], quat.shape[:-1]):
        raise ValueError(
            f"Expected `axes` to match number of rotations, got {axes.shape} axes "
            f"and {quat.shape} rotations."
        )

    # normalize axes
    axes = axes / xp_vector_norm(axes, axis=-1, keepdims=True, xp=xp)
    vdot_ax0_ax1 = xp.vecdot(axes[..., 0, :], axes[..., 1, :])
    vdot_ax1_ax2 = xp.vecdot(axes[..., 1, :], axes[..., 2, :])
    is_invalid = (vdot_ax0_ax1 >= 1e-7) | (vdot_ax1_ax2 >= 1e-7)
    if is_lazy_array(is_invalid):
        axes = xp.where(is_invalid[..., None, None], xp.nan, axes)
    elif xp.any(is_invalid):
        raise ValueError("Consecutive axes must be orthogonal.")

    angles = _compute_davenport_from_quat(
        quat,
        axes[..., 0, :],
        axes[..., 1, :],
        axes[..., 2, :],
        extrinsic,
        suppress_warnings,
    )
    if degrees:
        angles = _rad2deg(angles)
    return angles


def inv(quat: Array) -> Array:
    return xpx.at(quat)[..., :3].multiply(-1, copy=True)


def magnitude(quat: Array) -> Array:
    xp = array_namespace(quat)
    sin_q = xp_vector_norm(quat[..., :3], axis=-1, xp=xp)
    cos_q = xp.abs(quat[..., 3])
    angles = 2 * xp.atan2(sin_q, cos_q)
    return angles


def approx_equal(
    quat: Array, other: Array, atol: float | None = None, degrees: bool = False
) -> Array:
    if atol is None:
        if degrees:
            warnings.warn(
                "atol must be set to use the degrees flag, defaulting to 1e-8 radians.",
                stacklevel=2,
            )
        atol = 1e-8
    elif degrees:
        atol = _deg2rad(atol)

    if not broadcastable(quat.shape, other.shape):
        raise ValueError(
            f"Expected broadcastable shapes in both rotations, got {quat.shape[:-1]} "
            f"rotations in first and {other.shape[:-1]} rotations in second object."
        )

    quat_result = compose_quat(other, inv(quat))
    angles = magnitude(quat_result)
    return angles < atol


def mean(
    quat: Array,
    weights: ArrayLike | None = None,
    axis: None | int | tuple[int, ...] = None,
) -> Array:
    xp = array_namespace(quat)
    device = xp_device(quat)
    dtype = xp_result_type(quat, force_floating=True, xp=xp)
    if quat.shape[0] == 0:
        raise ValueError("Mean of an empty rotation set is undefined.")
    # Axis logic: For None, we reduce over all axes. For int, we only reduce over that
    # axis. For tuple, we reduce over all specified axes.
    all_axes = tuple(range(quat.ndim - 1))
    if axis is None:
        axis = all_axes
    elif isinstance(axis, int):
        axis = (axis,)
    if not isinstance(axis, tuple):
        raise ValueError("`axis` must be None, int, or tuple of ints.")
    # Ensure all axes are within bounds
    if axis != () and (min(axis) < -(quat.ndim - 1) or max(axis) > (quat.ndim - 2)):
        raise ValueError(
            f"axis {axis} is out of bounds for rotation with shape {quat.shape[:-1]}."
        )
    # Ensure all axes are positive and unique
    axis = tuple(sorted(set(x % (quat.ndim - 1) for x in axis)))

    lazy = is_lazy_array(quat)
    # Branching code is okay for checks that include meta info such as shapes and types
    quat_expand = quat[..., None, :]
    if weights is None:
        K = xp.matrix_transpose(quat_expand) @ quat_expand
    else:
        weights = xp.asarray(weights, dtype=dtype, device=device)
        neg_weights = weights < 0
        if not lazy and xp.any(neg_weights):
            raise ValueError("`weights` must be non-negative.")
        elif lazy:
            # We cannot check for negative weights because jit code needs to be
            # non-branching. We return NaN instead
            weights = xp.where(neg_weights, xp.nan, weights)

        if not broadcastable(quat.shape[:-1], weights.shape):
            raise ValueError(
                "Expected `weights` to be broadcastable to rotation shape, got shape "
                f"{weights.shape} for {quat.shape[:-1]} rotations."
            )

        # Make sure we can transpose quat
        weighted_quat = weights[..., None, None] * quat_expand
        K = xp.matrix_transpose(weighted_quat) @ quat_expand

    # Move reduction axes to the end
    keep_axes = tuple(i for i in all_axes if i not in axis)
    axes_order = keep_axes + axis
    K_reordered = xp.moveaxis(K, axes_order, all_axes)
    # Reshape to flatten reduction axes
    new_shape = K_reordered.shape[: len(keep_axes)] + (-1, 4, 4)
    K = xp.mean(xp.reshape(K_reordered, new_shape), axis=-3)
    _, v = xp.linalg.eigh(K)
    return v[..., -1]


def reduce(
    quat: Array,
    left: Array | None = None,
    right: Array | None = None,
) -> tuple[Array, Array | None, Array | None]:
    if left is None and right is None:
        return quat, None, None
    # DECISION: We cannot have variable number of return arguments for jit compiled
    # functions. We therefore always return the indices, and filter out later.
    # TOOD: Properly support broadcasting.
    xp = array_namespace(quat)
    quat = xpx.atleast_nd(quat, ndim=2, xp=xp)
    if left is None:
        left = xp.ones_like(quat)
    if right is None:
        right = xp.ones_like(quat)

    # We want to calculate the real components of q = l * p * r. It can
    # be shown that:
    #     qs = ls * ps * rs - ls * dot(pv, rv) - ps * dot(lv, rv)
    #          - rs * dot(lv, pv) - dot(cross(lv, pv), rv)
    # where ls and lv denote the scalar and vector components of l.

    p = quat
    ps, pv = _split_rotation(p, xp)
    ls, lv = _split_rotation(left, xp)
    rs, rv = _split_rotation(right, xp)

    # Compute each term without einsum (not accessible in the Array API)
    # First term: np.einsum("i,j,k", ls, ps, rs)
    term1 = ls[..., :, None, None] * ps[..., None, :, None] * rs[..., None, None, :]
    # Second term: np.einsum('i,jx,kx', ls, pv, rv)
    prv = xp.sum(pv[..., :, None, :] * rv[..., None, :, :], axis=-1)
    term2 = ls[..., :, None, None] * prv[..., None, :, :]
    # Third term: np.einsum('ix,j,kx', lv, ps, rv)
    lrv = xp.sum(lv[..., :, None, :] * rv[..., None, :, :], axis=-1)
    term3 = ps[..., None, :, None] * lrv[..., :, None, :]
    # Fourth term: np.einsum('ix,jx,k', lv, pv, rs)
    lpv = xp.sum(lv[..., :, None, :] * pv[..., None, :, :], axis=-1)
    term4 = rs[..., None, None, :] * lpv[..., :, :, None]
    # Fifth term: np.einsum('xyz,ix,jy,kz', e, lv, pv, rv). We want to avoid expanding
    # the einsum into a 6D tensor to avoid excessive memory usage. Instead, we compute
    # the cross product between lv and pv and then compute the dot product with rv.
    # First compute cross products between lv and pv
    lv_expanded = lv[..., :, None, :]
    pv_expanded = pv[..., None, :, :]
    cross_lp = xp.linalg.cross(lv_expanded, pv_expanded)
    # Then compute dot product with rv
    term5 = xp.sum(cross_lp[..., :, :, None, :] * rv[..., None, None, :, :], axis=-1)
    # Combine all terms with proper shape alignment
    qs = xp.abs(term1 - term2 - term3 - term4 - term5)
    qs = xp.reshape(xp.moveaxis(qs, 1, 0), (qs.shape[1], -1))

    # Find best indices from scalar components
    max_ind = xp.argmax(xp.reshape(qs, (qs.shape[0], -1)), axis=1)
    left_best = max_ind // rv.shape[0]
    right_best = max_ind % rv.shape[0]
    # Array API limitation: Integer index arrays are only allowed with integer indices
    # TODO: Can we somehow avoid this?
    all_idx = xp.reshape(xp.arange(left.shape[-1]), (1, -1))
    left_idx = xp.reshape(left_best, (-1, 1))
    left = left[left_idx, all_idx]
    right_idx = xp.reshape(right_best, (-1, 1))
    right = right[right_idx, all_idx]

    # Reduce the rotation using the best indices
    reduced = compose_quat(left, compose_quat(p, right))

    if left is None:
        left_best = None
    if right is None:
        right_best = None
    return reduced, left_best, right_best


def apply(quat: Array, points: Array, inverse: bool = False) -> Array:
    xp = array_namespace(quat)
    mat = as_matrix(quat)
    # We do not have access to einsum. To avoid broadcasting issues, we add a singleton
    # dimension to the points array and remove it after the operation.
    points = points[..., None]
    if not broadcastable(mat.shape, points.shape):
        raise ValueError(
            f"Cannot broadcast {quat.shape[:-1]} rotations to {points.shape[:-1]} "
            "vectors."
        )
    if inverse:
        # TODO: Replace with .mT once numpy 2.0 is the minimum supported version
        return (xp.matrix_transpose(mat) @ points)[..., 0]
    return (mat @ points)[..., 0]


def setitem(
    quat: Array, value: Array, indexer: int | slice | EllipsisType | None
) -> Array:
    return xpx.at(quat)[indexer, ...].set(value)


def align_vectors(
    a: Array, b: Array, weights: Array | None = None, return_sensitivity: bool = False
) -> tuple[Array, Array, Array]:
    xp = array_namespace(a)
    # Check input vectors
    dtype = xp_result_type(a, b, force_floating=True, xp=xp)
    a_original = xp.asarray(a, dtype=dtype)
    b_original = xp.asarray(b, dtype=dtype)
    a = xpx.atleast_nd(a_original, ndim=2, xp=xp)
    b = xpx.atleast_nd(b_original, ndim=2, xp=xp)
    if a.shape[-1] != 3:
        raise ValueError(
            f"Expected input `a` to have shape (3,) or (N, 3), got {a_original.shape}"
        )
    if b.shape[-1] != 3:
        raise ValueError(
            f"Expected input `b` to have shape (3,) or (N, 3), got {b_original.shape}"
        )
    if a.shape != b.shape:
        raise ValueError(
            "Expected inputs `a` and `b` to have same shapes"
            f", got {a_original.shape} and {b_original.shape} respectively."
        )
    if a.ndim > 2 or b.ndim > 2:  # This function does not support broadcasting
        raise ValueError(
            "Expected inputs `a` and `b` to have shape (3,) or (N, 3), got "
            f"{a_original.shape} and {b_original.shape} respectively."
        )
    N = a.shape[0]

    # Check weights
    if weights is None:
        weights = xp.ones(N, device=xp_device(a), dtype=a.dtype)
    else:
        weights = xp.asarray(weights, device=xp_device(a), dtype=a.dtype)
        if weights.ndim != 1:
            raise ValueError(
                f"Expected `weights` to be 1 dimensional, got shape {weights.shape}."
            )
        if N > 1 and (weights.shape[0] != N):
            raise ValueError(
                "Expected `weights` to have number of values equal to number of input "
                f"vectors, got {weights.shape[0]} values and {N} vectors."
            )
        # We can only check for negative weights in eager execution models. Lazy
        # backends will return NaNs instead.
        negative_weights = weights < 0
        if not is_lazy_array(negative_weights) and xp.any(negative_weights):
            raise ValueError("`weights` may not contain negative values")
        weights = xp.where(negative_weights, xp.nan, weights)

    # For the special case of a single vector pair, we use the infinite
    # weight code path
    weight_is_inf = xp.asarray([True]) if N == 1 else weights == xp.inf
    n_inf = xp.sum(xp.astype(weight_is_inf, a.dtype))
    # We can only error out on multiple infinite weights or sensitivity return with
    # infinite weights in eager execution models. Lazy backends will return NaNs.
    if not is_lazy_array(n_inf):
        if n_inf > 1:
            raise ValueError("Only one infinite weight is allowed")
        if n_inf == 1 and return_sensitivity:
            raise ValueError(
                "Cannot return sensitivity matrix with an "
                "infinite weight or one vector pair"
            )

    weights = xp.where(n_inf > 1, xp.nan, weights)

    inf_branch = xp.any(weight_is_inf, axis=-1)
    # DECISION: We cannot compute both branches for all frameworks. There are two main
    # reasons:
    # 1. Computing both for eager execution models is expensive.
    # 2. Some operations will fail when running the unused branch because of numerical
    # and algorithmical issues. Numpy e.g. will raise an exception when trying to
    # compute the svd of a matrix with infinite weights. To prevent this, we only
    # compute the branch that is needed. Lazy backends however require us to take the
    # full compute graph. Therefore, we use xp.where for lazy backends and a branching
    # version for eager frameworks.
    #
    # Note that we could also solve this by exploiting the externals of xpx.apply_where.
    # However, we'd have to rely on the implementation details of apply_where, which is
    # something we should avoid.
    # See https://github.com/scipy/scipy/pull/22777#discussion_r2028868364
    if is_lazy_array(inf_branch):
        q_opt, rssd, sensitivity = _align_vectors(a, b, weights)
        q_opt_inf, rssd_inf, sensitivity_inf = _align_vectors_fixed(a, b, weights)
        q_opt = xp.where(inf_branch, q_opt_inf, q_opt)
        rssd = xp.where(inf_branch, rssd_inf, rssd)
        sensitivity = xp.where(inf_branch, sensitivity_inf, sensitivity)
    else:
        if xp.any(inf_branch):
            q_opt, rssd, sensitivity = _align_vectors_fixed(a, b, weights)
        else:
            q_opt, rssd, sensitivity = _align_vectors(a, b, weights)
    return q_opt, rssd, sensitivity


def _align_vectors(a: Array, b: Array, weights: Array) -> tuple[Array, Array, Array]:
    xp = array_namespace(a)
    device = xp_device(a)
    B = (weights[:, None] * a).mT @ b
    u, s, vh = xp.linalg.svd(B)

    # Correct improper rotation if necessary (as in Kabsch algorithm)
    neg_det = xp.linalg.det(u @ vh) < 0
    s = xpx.at(s)[..., -1].set(xp.where(neg_det, -s[..., -1], s[..., -1]))
    u = xpx.at(u)[..., :, -1].set(xp.where(neg_det, -u[..., :, -1], u[..., :, -1]))

    C = u @ vh

    # DECISION: We cannot branch on the condition because jit code needs to be
    # non-branching. Hence, we omit the check for uniqueness
    # (s[1] + s[2] < 1e-16 * s[0])
    ssd = xp.sum(weights * xp.sum(b**2 + a**2, axis=-1), axis=-1) - 2 * xp.sum(
        s, axis=-1
    )
    rssd = xp.sqrt(xp.maximum(ssd, xp.zeros(1, device=device)))[..., 0]

    # TODO: We currently need to always compute the sensitivity matrix because lazy code
    # needs to be non-branching. We should check if compilers can optimize the where
    # statement (e.g. in jax) and check if we can have an eager version that only
    # evaluates the branch that is needed.
    # See xpx.apply_where, issue: https://github.com/data-apis/array-api-extra/pull/141
    zeta = (s[..., 0] + s[..., 1]) * (s[..., 1] + s[..., 2]) * (s[..., 2] + s[..., 0])
    kappa = s[..., 0] * s[..., 1] + s[..., 1] * s[..., 2] + s[..., 2] * s[..., 0]
    eye = xp.eye(3, dtype=a.dtype, device=device)
    sensitivity = xp.mean(weights) / zeta * (kappa * eye + B @ B.mT)
    q_opt = _from_matrix_orthogonal(C)
    return q_opt, rssd, sensitivity


def _align_vectors_fixed(
    a: Array, b: Array, weights: Array
) -> tuple[Array, Array, Array]:
    xp = array_namespace(a)
    device = xp_device(a)
    N = a.shape[0]
    weight_is_inf = xp.asarray([True], device=device) if N == 1 else weights == xp.inf
    # We cannot use boolean masks for indexing because of jax. For the same reason, we
    # also cannot use dynamic slices. As a workaround, we roll the array so that the
    # infinitely weighted vector is at index 0. We then use static slices to get the
    # primary and secondary vectors.
    #
    # Note that argmax fulfils a secondary purpose here:
    # When we trace this function with jax, this function might get executed even if
    # weight_is_inf does not have a single valid entry. Argmax will still give us a
    # valid index which allows us to proceed with the function (the results are
    # discarded anyways), whereas boolean indexing would result in invalid, zero-shaped
    # arrays.

    inf_idx = xp.argmax(xp.astype(weight_is_inf, xp.uint8))
    # xp.argmax returns an Array, but the specification of xp.roll does not support
    # Arrays as shifts. For lazy execution models we cannot convert to int because this
    # raises a concretization error. However, jax does accept Arrays as shifts which
    # allows jit compiling the function. Hence we do not convert to int for jax.
    # See https://github.com/data-apis/array-api/issues/914#issuecomment-2918959918
    if not is_jax(xp):
        inf_idx = int(inf_idx)
    a_sorted = xp.roll(a, shift=-inf_idx, axis=0)
    b_sorted = xp.roll(b, shift=-inf_idx, axis=0)
    weights_sorted = xp.roll(weights, shift=-inf_idx, axis=0)

    a_pri = a_sorted[0, ...][None, ...]  # Effectively [[0], ...]
    b_pri = b_sorted[0, ...][None, ...]
    a_pri_norm = xp_vector_norm(a_pri, axis=-1, keepdims=True, xp=xp)
    b_pri_norm = xp_vector_norm(b_pri, axis=-1, keepdims=True, xp=xp)
    if not is_lazy_array(a_pri_norm):
        if xp.any(a_pri_norm == 0) or xp.any(b_pri_norm == 0):
            raise ValueError("Cannot align zero length primary vectors")

    # We cannot error out on zero length vectors. We set the norm to NaN to avoid
    # division by zero and mark the result as invalid.
    a_pri_norm = xpx.at(a_pri_norm)[a_pri_norm == 0].set(xp.nan)
    b_pri_norm = xpx.at(b_pri_norm)[b_pri_norm == 0].set(xp.nan)

    a_pri = a_pri / a_pri_norm
    b_pri = b_pri / b_pri_norm

    # We first find the minimum angle rotation between the primary
    # vectors.
    cross = xp.linalg.cross(b_pri[..., 0, :], a_pri[..., 0, :])
    cross_norm = xp_vector_norm(cross, axis=-1, xp=xp)
    theta = xp.atan2(cross_norm, xp.sum(a_pri[..., 0, :] * b_pri[..., 0, :], axis=-1))
    tolerance = 1e-3  # tolerance for small angle approximation (rad)
    q_flip = xp.asarray([0.0, 0.0, 0.0, 1.0], dtype=a_pri.dtype, device=device)

    # Near pi radians, the Taylor series approximation of x/sin(x) diverges, so for
    # numerical stability we flip pi and then rotate back by the small angle pi - theta
    flip = xp.pi - theta < tolerance
    # For antiparallel vectors, cross = [0, 0, 0] so we need to manually set an
    # arbitrary orthogonal axis of rotation
    i = xp.argmin(xp.abs(a_pri[..., 0, :]))
    # TODO: Array API does not support fancy indexing __setitem__. The code is
    # equivalent to doing the following:
    # r_components = xp.zeros(3)
    # r_components = xpx.at(r_components)[i - 1].set(a_pri[0, i - 2])
    # r_components = xpx.at(r_components)[i - 2].set(-a_pri[0, i - 1])
    r_components = xp.stack(
        [
            xp.where(i == 1, a_pri[0, 2], xp.where(i == 2, -a_pri[0, 1], 0)),
            xp.where(i == 2, a_pri[0, 0], xp.where(i == 0, -a_pri[0, 2], 0)),
            xp.where(i == 0, a_pri[0, 1], xp.where(i == 1, -a_pri[0, 0], 0)),
        ]
    )
    r = xp.where(cross_norm == 0, r_components, cross)

    q_flip = xp.where(flip, from_rotvec(r / xp_vector_norm(r, xp=xp) * xp.pi), q_flip)
    theta = xp.where(flip, xp.pi - theta, theta)
    cross = xp.where(flip, -cross, cross)

    # Small angle Taylor series approximation for numerical stability
    theta2 = theta * theta
    small_scale = xp.abs(theta) < tolerance
    r_small_scale = cross * (1 + theta2 / 6 + theta2 * theta2 * 7 / 360)
    # We need to handle the case where theta is 0 to avoid division by zero. We use the
    # value of the Taylor series approximation, but non-branching operations require
    # that we still divide by the angle. Since we do not use the result where the angle
    # is close to 0, this is safe.
    theta = theta + xp.asarray(small_scale, dtype=theta.dtype)
    r_large_scale = cross * theta / xp.sin(theta)
    r = xp.where(small_scale, r_small_scale, r_large_scale)
    q_pri = compose_quat(from_rotvec(r), q_flip)

    # Case 1): No secondary vectors, q_opt is q_pri -> Immediately done
    # Case 2): Secondary vectors exist
    # We cannot use boolean masks here because of jax
    a_sec = a_sorted[1:, ...]
    b_sec = b_sorted[1:, ...]
    weights_sec = weights_sorted[1:]

    # We apply the first rotation to the b vectors to align the
    # primary vectors, resulting in vectors c. The secondary
    # vectors must now be rotated about that primary vector to best
    # align c to a.
    c_sec = apply(q_pri, b_sec)

    # Calculate vector components for the angle calculation. The
    # angle phi to rotate a single 2D vector C to align to 2D
    # vector A in the xy plane can be found with the equation
    # phi = atan2(cross(C, A), dot(C, A))
    #     = atan2(|C|*|A|*sin(phi), |C|*|A|*cos(phi))
    # The below equations perform the same operation, but with the
    # 3D rotation restricted to the 2D plane normal to a_pri, where
    # the secondary vectors are projected into that plane. We then
    # find the composite angle from the weighted sum of the
    # axial components in that plane, minimizing the 2D alignment
    # problem.
    sin_term = xp.linalg.vecdot(xp.linalg.cross(c_sec, a_sec), a_pri)
    cos_term = xp.linalg.vecdot(c_sec, a_sec) - (
        xp.linalg.vecdot(c_sec, a_pri) * xp.linalg.vecdot(a_sec, a_pri)
    )
    phi = xp.atan2(xp.sum(weights_sec * sin_term), xp.sum(weights_sec * cos_term))
    q_sec = from_rotvec(phi * a_pri[0, ...])

    # Compose these to get the optimal rotation
    q_opt = q_pri if N == 1 else compose_quat(q_sec, q_pri)

    # Calculate the root sum squared distance. We force the error to
    # be zero for the infinite weight vectors since they will align
    # exactly.
    mask = ((N > 1) | (weights[0] == xp.inf)) & weight_is_inf
    # Skip non-infinite weight single vectors pairs, we used the
    # infinite weight code path but don't want to zero that weight
    weights_inf_zero = xpx.at(weights, mask).set(0, copy=True, xp=xp)
    a_est = apply(q_opt, b)
    rssd = xp.sqrt(xp.sum(weights_inf_zero @ (a - a_est) ** 2))

    mask = xp.any(xp.isnan(weights), axis=-1)
    q_opt = xp.where(mask, xp.nan, q_opt)
    return q_opt, rssd, xp.asarray(xp.nan, device=device)


def pow(quat: Array, n: float | Array) -> Array:
    xp = array_namespace(quat)
    device = xp_device(quat)
    # If n is an array, we sanitize it to a scalar and promote quat and n to
    # the same dtype.
    if is_array_api_obj(n):
        if n.shape == (1,):  # pyrefly:ignore[missing-attribute]
            n = n[0]  # pyrefly:ignore[bad-index]
        elif n.ndim != 0:  # pyrefly:ignore[missing-attribute]
            raise ValueError("Array exponent must be a scalar")
        quat, n = xp_promote(quat, n, force_floating=True, xp=xp)

    # If n is a lazy array, we cannot take fast paths for special cases.
    if is_lazy_array(n):
        result = from_rotvec(n * as_rotvec(quat))  # general scaling of rotation angle
        # Special cases 0 -> identity, -1 -> inv, 1 -> copy
        identity = xp.zeros((*quat.shape[:-1], 4), dtype=quat.dtype, device=device)
        identity = xpx.at(identity)[..., 3].set(1)
        result = xp.where(n == 0, identity, result)
        result = xp.where(n == -1, inv(quat), result)
        result = xp.where(n == 1, quat, result)
        return result
    if n == 0:
        identity = xp.zeros((*quat.shape[:-1], 4), dtype=quat.dtype, device=device)
        return xpx.at(identity)[..., 3].set(1)
    if n == -1:
        return inv(quat)
    if n == 1:
        return quat
    return from_rotvec(n * as_rotvec(quat))


def _normalize_quaternion(quat: Array) -> Array:
    xp = array_namespace(quat)
    quat_norm = xp_vector_norm(quat, axis=-1, keepdims=True, xp=xp)
    zero_norm = quat_norm == 0
    if is_lazy_array(quat_norm):
        quat = xp.where(zero_norm, xp.nan, quat)
    elif xp.any(zero_norm):
        raise ValueError("Found zero norm quaternions in `quat`.")
    return quat / quat_norm


def _quat_canonical(quat: Array) -> Array:
    xp = array_namespace(quat)
    mask = quat[..., 3] < 0
    zero_w = quat[..., 3] == 0
    mask = mask | (zero_w & (quat[..., 0] < 0))
    zero_wx = zero_w & (quat[..., 0] == 0)
    mask = mask | (zero_wx & (quat[..., 1] < 0))
    zero_wxy = zero_wx & (quat[..., 1] == 0)
    mask = mask | (zero_wxy & (quat[..., 2] < 0))
    return xp.where(mask[..., None], -quat, quat)


def _elementary_basis_index(axis: str) -> int:
    if axis == "x":
        return 0
    elif axis == "y":
        return 1
    elif axis == "z":
        return 2
    raise ValueError(f"Expected axis to be from ['x', 'y', 'z'], got {axis}")


def _compute_davenport_from_quat(
    quat: Array,
    n1: Array,
    n2: Array,
    n3: Array,
    extrinsic: bool,
    suppress_warnings: bool,
) -> Array:
    # The algorithm assumes extrinsic frame transformations. The algorithm
    # in the paper is formulated for rotation quaternions, which are stored
    # directly by Rotation.
    # Adapt the algorithm for our case by reversing both axis sequence and
    # angles for intrinsic rotations when needed
    xp = array_namespace(quat)
    n1, n3 = (n1, n3) if extrinsic else (n3, n1)

    n_cross = xp.linalg.cross(n1, n2)
    lamb = xp.atan2(xp.vecdot(n3, n_cross), xp.vecdot(n3, n1))

    # alternative set of angles compatible with as_euler implementation
    mask = lamb < 0
    n2 = xp.where(mask[..., None], -n2, n2)
    lamb = xp.where(mask, -lamb, lamb)
    n_cross = xp.where(mask[..., None], -n_cross, n_cross)
    correct_set = mask

    quat_lamb = xp.concat(
        (xp.sin(lamb / 2)[..., None] * n2, xp.cos(lamb / 2)[..., None]), axis=-1
    )

    q_trans = compose_quat(quat_lamb, quat)
    a = q_trans[..., 3]
    b = xp.linalg.vecdot(q_trans[..., :3], n1)
    c = xp.linalg.vecdot(q_trans[..., :3], n2)
    d = xp.linalg.vecdot(q_trans[..., :3], n_cross)

    angles = _get_angles(extrinsic, False, 1, lamb, a, b, c, d, suppress_warnings)
    angles = xpx.at(angles)[..., 1].set(
        xp.where(correct_set, -angles[..., 1], angles[..., 1])
    )

    return angles


def _elementary_quat_compose(axes: list[int], angles: Array, intrinsic: bool) -> Array:
    xp = array_namespace(angles)
    device = xp_device(angles)
    quat = _make_elementary_quat(axes[0], angles[..., 0], device=device, xp=xp)
    for i in range(1, len(axes)):
        ax_quat = _make_elementary_quat(axes[i], angles[..., i], device=device, xp=xp)
        quat = compose_quat(quat, ax_quat) if intrinsic else compose_quat(ax_quat, quat)
    return quat


def _make_elementary_quat(axis: int, angle: Array, device, xp) -> Array:
    quat = xp.zeros((*angle.shape, 4), dtype=angle.dtype, device=device)
    quat = xpx.at(quat)[..., 3].set(xp.cos(angle / 2.0))
    quat = xpx.at(quat)[..., axis].set(xp.sin(angle / 2.0))
    return quat


def _get_angles(
    extrinsic: bool,
    symmetric: bool,
    sign: int,
    lamb: float,
    a: Array,
    b: Array,
    c: Array,
    d: Array,
    suppress_warnings: bool,
) -> Array:
    xp = array_namespace(a)
    device = xp_device(a)
    eps = 1e-7
    half_sum = xp.atan2(b, a)
    half_diff = xp.atan2(d, c)
    # We zero-initialize to automatically cover singular cases where the second angle is
    # not defined uniquely.
    angles = xp.zeros((*a.shape, 3), dtype=a.dtype, device=device)

    angles = xpx.at(angles)[..., 1].set(2 * xp.atan2(xp.hypot(c, d), xp.hypot(a, b)))

    angle_first = 0 if extrinsic else 2
    angle_third = 2 if extrinsic else 0

    # Check if the second angle is close to 0 or pi, causing a singularity.
    # - Case 0: Second angle is neither close to 0 nor pi.
    # - Case 1: Second angle is close to 0.
    # - Case 2: Second angle is close to pi.
    case1 = xp.abs(angles[..., 1]) <= eps
    case2 = xp.abs(angles[..., 1] - xp.pi) <= eps
    case0 = ~(case1 | case2)
    if not suppress_warnings and not is_lazy_array(case0) and xp.any(~case0):
        warnings.warn(
            "Gimbal lock detected. Setting third angle to zero "
            "since it is not possible to uniquely determine "
            "all angles.",
            stacklevel=3,
        )

    # This writes case1 into a0 where True and case2 everywhere else. This is sound
    # because we later overwrite any values without singularity with the regular value
    # of case0. The second angle is covered by default since we zero-initialized the
    # second dimension.
    a0 = xp.where(case1, 2 * half_sum, 2 * half_diff * (-1 if extrinsic else 1))
    angles = xpx.at(angles)[..., 0].set(a0)

    # We overwrite the values of angles without singularities (case0)
    a1 = xp.where(case0, half_sum - half_diff, angles[..., angle_first])
    angles = xpx.at(angles)[..., angle_first].set(a1)

    # Same as above but for the third angle. We overwrite the non-singular case0 values
    a3 = xp.where(case0, half_sum + half_diff, angles[..., angle_third])
    if not symmetric:
        a3 = a3 * sign
        angles = xpx.at(angles)[..., 1].set(angles[..., 1] - lamb)
    angles = xpx.at(angles)[..., angle_third].set(a3)

    angles = (angles + xp.pi) % (2 * xp.pi) - xp.pi
    return angles


def compose_quat(p: Array, q: Array) -> Array:
    xp = array_namespace(p)
    cross = xp.linalg.cross(p[..., :3], q[..., :3])
    qx = p[..., 3] * q[..., 0] + q[..., 3] * p[..., 0] + cross[..., 0]
    qy = p[..., 3] * q[..., 1] + q[..., 3] * p[..., 1] + cross[..., 1]
    qz = p[..., 3] * q[..., 2] + q[..., 3] * p[..., 2] + cross[..., 2]
    qw = (
        p[..., 3] * q[..., 3]
        - p[..., 0] * q[..., 0]
        - p[..., 1] * q[..., 1]
        - p[..., 2] * q[..., 2]
    )
    quat = xp.stack([qx, qy, qz, qw], axis=-1)
    return quat


def _split_rotation(q: Array, xp) -> tuple[Array, Array]:
    q = xpx.atleast_nd(q, ndim=2, xp=xp)
    return q[..., -1], q[..., :-1]


def _deg2rad(angles: Array) -> Array:
    return angles * (np.pi / 180.0)


def _rad2deg(angles: Array) -> Array:
    return angles * (180.0 / np.pi)



from types import GenericAlias
import numpy as np


def check_arguments(fun, y0, support_complex):
    """Helper function for checking arguments common to all solvers."""
    y0 = np.asarray(y0)
    if np.issubdtype(y0.dtype, np.complexfloating):
        if not support_complex:
            raise ValueError("`y0` is complex, but the chosen solver does "
                             "not support integration in a complex domain.")
        dtype = complex
    else:
        dtype = float
    y0 = y0.astype(dtype, copy=False)

    if y0.ndim != 1:
        raise ValueError("`y0` must be 1-dimensional.")

    if not np.isfinite(y0).all():
        raise ValueError("All components of the initial state `y0` must be finite.")

    def fun_wrapped(t, y):
        return np.asarray(fun(t, y), dtype=dtype)

    return fun_wrapped, y0


class OdeSolver:
    """Base class for ODE solvers.

    In order to implement a new solver you need to follow the guidelines:

    1. A constructor must accept parameters presented in the base class
       (listed below) along with any other parameters specific to a solver.
    2. A constructor must accept arbitrary extraneous arguments
       ``**extraneous``, but warn that these arguments are irrelevant
       using `common.warn_extraneous` function. Do not pass these
       arguments to the base class.
    3. A solver must implement a private method ``_step_impl(self)`` which
       propagates a solver one step further. It must return tuple
       ``(success, message)``, where ``success`` is a boolean indicating
       whether a step was successful, and ``message`` is a string
       containing description of a failure if a step failed or None
       otherwise.
    4. A solver must implement a private method ``_dense_output_impl(self)``,
       which returns a `DenseOutput` object covering the last successful
       step.
    5. A solver must have attributes listed below in Attributes section.
       Note that ``t_old`` and ``step_size`` are updated automatically.
    6. Use ``fun(self, t, y)`` method for the system rhs evaluation, this
       way the number of function evaluations (`nfev`) will be tracked
       automatically.
    7. For convenience, a base class provides ``fun_single(self, t, y)`` and
       ``fun_vectorized(self, t, y)`` for evaluating the rhs in
       non-vectorized and vectorized fashions respectively (regardless of
       how `fun` from the constructor is implemented). These calls don't
       increment `nfev`.
    8. If a solver uses a Jacobian matrix and LU decompositions, it should
       track the number of Jacobian evaluations (`njev`) and the number of
       LU decompositions (`nlu`).
    9. By convention, the function evaluations used to compute a finite
       difference approximation of the Jacobian should not be counted in
       `nfev`, thus use ``fun_single(self, t, y)`` or
       ``fun_vectorized(self, t, y)`` when computing a finite difference
       approximation of the Jacobian.

    Parameters
    ----------
    fun : callable
        Right-hand side of the system: the time derivative of the state ``y``
        at time ``t``. The calling signature is ``fun(t, y)``, where ``t`` is a
        scalar and ``y`` is an ndarray with ``len(y) = len(y0)``. ``fun`` must
        return an array of the same shape as ``y``. See `vectorized` for more
        information.
    t0 : float
        Initial time.
    y0 : array_like, shape (n,)
        Initial state.
    t_bound : float
        Boundary time --- the integration won't continue beyond it. It also
        determines the direction of the integration.
    vectorized : bool
        Whether `fun` can be called in a vectorized fashion. Default is False.

        If ``vectorized`` is False, `fun` will always be called with ``y`` of
        shape ``(n,)``, where ``n = len(y0)``.

        If ``vectorized`` is True, `fun` may be called with ``y`` of shape
        ``(n, k)``, where ``k`` is an integer. In this case, `fun` must behave
        such that ``fun(t, y)[:, i] == fun(t, y[:, i])`` (i.e. each column of
        the returned array is the time derivative of the state corresponding
        with a column of ``y``).

        Setting ``vectorized=True`` allows for faster finite difference
        approximation of the Jacobian by methods 'Radau' and 'BDF', but
        will result in slower execution for other methods. It can also
        result in slower overall execution for 'Radau' and 'BDF' in some
        circumstances (e.g. small ``len(y0)``).
    support_complex : bool, optional
        Whether integration in a complex domain should be supported.
        Generally determined by a derived solver class capabilities.
        Default is False.

    Attributes
    ----------
    n : int
        Number of equations.
    status : str
        Current status of the solver: 'running', 'finished' or 'failed'.
    t_bound : float
        Boundary time.
    direction : float
        Integration direction: +1 or -1.
    t : float
        Current time.
    y : ndarray
        Current state.
    t_old : float
        Previous time. None if no steps were made yet.
    step_size : float
        Size of the last successful step. None if no steps were made yet.
    nfev : int
        Number of the system's rhs evaluations.
    njev : int
        Number of the Jacobian evaluations.
    nlu : int
        Number of LU decompositions.
    """
    TOO_SMALL_STEP = "Required step size is less than spacing between numbers."

    # generic type compatibility with scipy-stubs
    __class_getitem__: classmethod = classmethod(GenericAlias)

    def __init__(self, fun, t0, y0, t_bound, vectorized,
                 support_complex=False):
        self.t_old = None
        self.t = t0
        self._fun, self.y = check_arguments(fun, y0, support_complex)
        self.t_bound = t_bound
        self.vectorized = vectorized

        if vectorized:
            def fun_single(t, y):
                return self._fun(t, y[:, None]).ravel()
            fun_vectorized = self._fun
        else:
            fun_single = self._fun

            def fun_vectorized(t, y):
                f = np.empty_like(y)
                for i, yi in enumerate(y.T):
                    f[:, i] = self._fun(t, yi)
                return f

        def fun(t, y):
            self.nfev += 1
            return self.fun_single(t, y)

        self.fun = fun
        self.fun_single = fun_single
        self.fun_vectorized = fun_vectorized

        self.direction = np.sign(t_bound - t0) if t_bound != t0 else 1
        self.n = self.y.size
        self.status = 'running'

        self.nfev = 0
        self.njev = 0
        self.nlu = 0

    @property
    def step_size(self):
        if self.t_old is None:
            return None
        else:
            return np.abs(self.t - self.t_old)

    def step(self):
        """Perform one integration step.

        Returns
        -------
        message : str or None
            Report from the solver. Typically a reason for a failure if
            `self.status` is 'failed' after the step was taken or None
            otherwise.
        """
        if self.status != 'running':
            raise RuntimeError("Attempt to step on a failed or finished "
                               "solver.")

        if self.n == 0 or self.t == self.t_bound:
            # Handle corner cases of empty solver or no integration.
            self.t_old = self.t
            self.t = self.t_bound
            message = None
            self.status = 'finished'
        else:
            t = self.t
            success, message = self._step_impl()

            if not success:
                self.status = 'failed'
            else:
                self.t_old = t
                if self.direction * (self.t - self.t_bound) >= 0:
                    self.status = 'finished'

        return message

    def dense_output(self):
        """Compute a local interpolant over the last successful step.

        Returns
        -------
        sol : `DenseOutput`
            Local interpolant over the last successful step.
        """
        if self.t_old is None:
            raise RuntimeError("Dense output is available after a successful "
                               "step was made.")

        if self.n == 0 or self.t == self.t_old:
            # Handle corner cases of empty solver and no integration.
            return ConstantDenseOutput(self.t_old, self.t, self.y)
        else:
            return self._dense_output_impl()

    def _step_impl(self):
        raise NotImplementedError

    def _dense_output_impl(self):
        raise NotImplementedError


class DenseOutput:
    """Base class for local interpolant over step made by an ODE solver.

    It interpolates between `t_min` and `t_max` (see Attributes below).
    Evaluation outside this interval is not forbidden, but the accuracy is not
    guaranteed.

    Parameters
    ----------
    t_old : float
        Previous time.
    t : float
        Current time.

    Attributes
    ----------
    t_min, t_max : float
        Time range of the interpolation.
    """

    # generic type compatibility with scipy-stubs
    __class_getitem__: classmethod = classmethod(GenericAlias)

    def __init__(self, t_old, t):
        self.t_old = t_old
        self.t = t
        self.t_min = min(t, t_old)
        self.t_max = max(t, t_old)

    def __call__(self, t):
        """Evaluate the interpolant.

        Parameters
        ----------
        t : float or array_like with shape (n_points,)
            Points to evaluate the solution at.

        Returns
        -------
        y : ndarray, shape (n,) or (n, n_points)
            Computed values. Shape depends on whether `t` was a scalar or a
            1-D array.
        """
        t = np.asarray(t)
        if t.ndim > 1:
            raise ValueError("`t` must be a float or a 1-D array.")
        return self._call_impl(t)

    def _call_impl(self, t):
        raise NotImplementedError


class ConstantDenseOutput(DenseOutput):
    """Constant value interpolator.

    This class used for degenerate integration cases: equal integration limits
    or a system with 0 equations.
    """
    def __init__(self, t_old, t, value):
        super().__init__(t_old, t)
        self.value = value

    def _call_impl(self, t):
        if t.ndim == 0:
            return self.value
        else:
            ret = np.empty((self.value.shape[0], t.shape[0]))
            ret[:] = self.value[:, None]
            return ret



from __future__ import annotations

from collections.abc import Sequence
from typing import Literal

import torch  # noqa: F401
import torch.fft

from ._typing import Array
from .._internal import clone_module

__all__ = clone_module("torch.fft", globals())

# Several torch fft functions do not map axes to dim

def fftn(
    x: Array,
    /,
    *,
    s: Sequence[int] = None,
    axes: Sequence[int] = None,
    norm: Literal["backward", "ortho", "forward"] = "backward",
    **kwargs: object,
) -> Array:
    return torch.fft.fftn(x, s=s, dim=axes, norm=norm, **kwargs)

def ifftn(
    x: Array,
    /,
    *,
    s: Sequence[int] = None,
    axes: Sequence[int] = None,
    norm: Literal["backward", "ortho", "forward"] = "backward",
    **kwargs: object,
) -> Array:
    return torch.fft.ifftn(x, s=s, dim=axes, norm=norm, **kwargs)

def rfftn(
    x: Array,
    /,
    *,
    s: Sequence[int] = None,
    axes: Sequence[int] = None,
    norm: Literal["backward", "ortho", "forward"] = "backward",
    **kwargs: object,
) -> Array:
    return torch.fft.rfftn(x, s=s, dim=axes, norm=norm, **kwargs)

def irfftn(
    x: Array,
    /,
    *,
    s: Sequence[int] = None,
    axes: Sequence[int] = None,
    norm: Literal["backward", "ortho", "forward"] = "backward",
    **kwargs: object,
) -> Array:
    return torch.fft.irfftn(x, s=s, dim=axes, norm=norm, **kwargs)

def fftshift(
    x: Array,
    /,
    *,
    axes: int | Sequence[int] = None,
    **kwargs: object,
) -> Array:
    return torch.fft.fftshift(x, dim=axes, **kwargs)

def ifftshift(
    x: Array,
    /,
    *,
    axes: int | Sequence[int] = None,
    **kwargs: object,
) -> Array:
    return torch.fft.ifftshift(x, dim=axes, **kwargs)


__all__ += ["fftn", "ifftn", "rfftn", "irfftn", "fftshift", "ifftshift"]

def __dir__() -> list[str]:
    return __all__
